# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *
from gitfourchette.toolbox.recolorsvgiconengine import RecolorSvgIconEngine


_stockIconCache: dict[int, QIcon] = {}
_stockIconHtmlCache: dict[int, str] = {}

# Override some icon IDs depending on desktop environment
_overrideIconIds: dict[str, str] = {}

_autoDarkVariants = {
    "achtung",
    "git-lfs",
    "git-lfs-add",
    "git-lfs-remove",
    "status_a",
    "status_c",  # just in case we ever add this one
    "status_d",
    "status_m",
    "status_r",
    "status_t",
    "status_x",
    # status_u intentionally omitted. Its white outline should stand out against a white background.
}


def _iconOverrideTable() -> dict[str, str]:
    overrides = {
        "status_?": "status_a",  # Use Added icon for Untracked
    }

    assert QApplication.instance(), "need app instance for QIcon.themeName()"
    iconTheme = QIcon.themeName().casefold()

    # Use native warning icon in all contexts on Mac & Windows
    if MACOS or WINDOWS:  # pragma: no cover
        overrides["achtung"] = "SP_MessageBoxWarning"

    # Override Ubuntu default theme's scary red icon for warnings
    if FREEDESKTOP and iconTheme.startswith("yaru"):  # pragma: no cover
        overrides["SP_MessageBoxWarning"] = "warning-small-symbolic"

    return overrides


def stockIcon(iconId: str, colorTable="") -> QIcon:
    # Special cases
    if not _overrideIconIds:
        _overrideIconIds.update(_iconOverrideTable())
        assert _overrideIconIds, "expecting overrides to contain at least one entry"
    iconId = _overrideIconIds.get(iconId, iconId)

    if RecolorSvgIconEngine.IconColors.preferDarkVariants and iconId in _autoDarkVariants:
        assert not colorTable
        colorTable = "white=#000 black=#fff"

    # Compute cache key
    key = hash(iconId) ^ hash(colorTable)

    # Attempt to get cached icon
    try:
        return _stockIconCache[key]
    except KeyError:
        pass

    # Find path to icon file (if any)
    iconPath = ""
    for ext in ".svg", ".png":
        file = QFile(f"assets:icons/{iconId}{ext}")
        if file.exists():
            iconPath = file.fileName()
            break

    # Create QIcon
    if iconPath.endswith(".svg"):
        # Dynamic SVG icon
        engine = RecolorSvgIconEngine(iconPath, colorTable)
        icon = QIcon(engine)
    elif iconPath:
        # Bitmap file
        icon = QIcon(iconPath)
    elif iconId.startswith("SP_"):
        # Qt standard pixmaps (with "SP_" prefix)
        entry = getattr(QStyle.StandardPixmap, iconId)
        icon = QApplication.style().standardIcon(entry)
    else:
        # Fall back to theme icon
        icon = QIcon.fromTheme(iconId)

    assert iconPath.endswith(".svg") or not colorTable, f"can't remap colors in non-SVG icon! {iconId}"

    # Cache icon
    _stockIconCache[key] = icon
    return icon


def stockIconImgTag(iconId: str, dpr: float = 0) -> str:
    # Qt's HTML subset renders <img> tags at devicePixelRatio=1. To avoid ugly
    # stretching beyond 1x, we'll pre-render the image at the desired dpr. This
    # makes most sense for SVG images which make up the bulk of our icon bank.
    if dpr == 0:
        # devicePixelRatio not specified, figure it out.
        app = QApplication.instance()
        try:
            # Use mainWindow's dpr
            window: QMainWindow = app.mainWindow  # type: ignore[attr-defined]
            dpr = window.devicePixelRatio()
        except AttributeError:
            # Too early, no window yet; fall back to highest dpr on the system.
            # Note: on Wayland, this may be higher than the actual dpr if
            # fractional scaling is enabled (e.g. this may return 2.0 if your
            # system is set up for 1.25 frac scaling)
            dpr = app.devicePixelRatio()  # type: ignore[attr-defined]  # incomplete stubs

    key = hash(iconId) ^ hash(dpr)

    try:
        return _stockIconHtmlCache[key]
    except KeyError:
        pass

    size = QSize(16, 16)

    if dpr == 1 or QT5:
        # Pre-rendering not necessary for dpr=1
        # (Qt 5: can't pass devicePixelRatio to QIcon.pixmap())
        src = f"assets:icons/{iconId}"
    else:
        # Pre-render at the given dpr into a temporary PNG file.
        icon = stockIcon(iconId)
        pixmap = icon.pixmap(size, dpr)
        fileName = f"i{int(dpr*100)}-{key & 0xFFFFFFFF:x}.png"
        src = f"{qTempDir()}/{fileName}"
        pixmap.save(src)

    tag = f"<img src='{src}' width={size.width()} height={size.height()} style='vertical-align: bottom;'/>"

    _stockIconHtmlCache[key] = tag
    return tag


def clearStockIconCache():
    _stockIconCache.clear()
    _stockIconHtmlCache.clear()

    # Force reevaluate color scheme
    RecolorSvgIconEngine.IconColors.initialized = False



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import base64
import dataclasses
import json
import logging
import os
import typing
from types import NoneType, UnionType, GenericAlias
from typing import Any

from gitfourchette.porcelain import *
from gitfourchette.qt import *

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class PrefsFile:
    _filename = ""
    _allowMakeDirs = True
    _fileVersionKey = "_version"

    _dirty = False
    _fileVersion = ""

    def getParentDir(self) -> str:
        if APP_TESTMODE:
            return os.path.join(qTempDir(), "testmode-config")
        return QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppConfigLocation)

    def _getFullPath(self, forWriting: bool) -> str:
        assert self._filename != "", "you must override _filename"

        prefsDir = self.getParentDir()
        if not prefsDir:
            return ""

        if forWriting:
            if self._allowMakeDirs or False:
                os.makedirs(prefsDir, exist_ok=True)
            elif not os.path.isdir(prefsDir):
                return ""

        fullPath = os.path.join(prefsDir, self._filename)

        if not forWriting and not os.path.isfile(fullPath):
            return ""

        return fullPath

    def setDirty(self):
        self._dirty = True

    def isDirty(self):
        return self._dirty

    def reset(self):
        assert dataclasses.is_dataclass(self)
        for f in dataclasses.fields(self):
            if f.default_factory != dataclasses.MISSING:
                obj = f.default_factory()
            else:
                obj = f.default
            self.__dict__[f.name] = obj

    def write(self, force=False) -> str:
        # Prepare the path
        prefsPath = self._getFullPath(forWriting=True)
        if not prefsPath:
            logger.warning("Couldn't get path for writing")
            return ""

        # Filter the values
        assert dataclasses.is_dataclass(self)
        fields: dict[str, dataclasses.Field] = self.__dataclass_fields__
        filtered = {}
        for key, field in fields.items():
            # Skip private fields
            if key.startswith("_"):
                continue

            if field.default_factory != dataclasses.MISSING:
                default = field.default_factory()
            else:
                default = field.default

            value = self.__dict__[key]

            # Skip default values
            if value == default:
                continue

            # Make the value JSON-friendly
            value = self.encode(value)
            filtered[key] = value

        # If the filtered object comes out empty (all defaults)
        # avoid cluttering the directory - don't write out an empty object
        if not filtered:
            if not self._getFullPath(forWriting=False):
                # File doesn't exist - don't write out an empty object
                logger.debug("Not writing empty object")
            else:
                logger.debug("Deleting prefs file because we want defaults")
                os.unlink(prefsPath)
            return ""

        # Inject file version into JSON object before writing
        self._fileVersion = APP_VERSION
        filtered[self._fileVersionKey] = self._fileVersion

        # Dump the object to disk
        with open(prefsPath, "w", encoding="utf-8") as jsonFile:
            json.dump(obj=filtered, fp=jsonFile, indent='\t')
        self._dirty = False

        logger.info(f"Wrote {prefsPath}")
        return prefsPath

    def load(self) -> bool:
        prefsPath = self._getFullPath(forWriting=False)
        if not prefsPath:  # couldn't be found
            return False

        # Load JSON blob
        with open(prefsPath, encoding="utf-8") as file:
            try:
                jsonObject: dict[str, Any] = json.load(file)
            except ValueError as loadError:
                logger.warning(f"{prefsPath}: {loadError}", exc_info=True)
                return False

        if not isinstance(jsonObject, dict):
            logger.warning(f"{prefsPath}: not a JSON dict")  # type: ignore[unreachable]
            return False

        # Pop file version before decoding user fields
        fv = jsonObject.pop(self._fileVersionKey, "")
        if isinstance(fv, str):
            self._fileVersion = fv

        assert dataclasses.is_dataclass(self)
        fields: dict[str, dataclasses.Field] = self.__dataclass_fields__

        # Decode values and store them in this object
        for key, value in jsonObject.items():
            if key.startswith('_') or key not in fields:
                logger.warning(f"{prefsPath}: dropping key: {key}")
                continue

            if value is None:
                continue

            try:
                value = self.decode(value, fields[key].type)  # type: ignore
            except ValueError as error:
                logger.warning(f"{prefsPath}: {key}: {error}")
                continue

            self.__dict__[key] = value

        self._dirty = False
        return True

    @staticmethod
    def encode(o: Any) -> Any:
        """ Encode a value to make it JSON-friendly """
        if isinstance(o, enum.Enum):  # Convert Qt enums to plain old data type
            return o.value
        elif type(o) is bytes:
            return base64.b64encode(o).decode("ascii")
        elif type(o) is set:
            return list(o)
        elif type(o) is Signature:
            return {"name": o.name, "email": o.email, "time": o.time, "offset": o.offset}
        return o

    @staticmethod
    def decode(o: Any, dstType: type | UnionType | GenericAlias) -> Any:
        """ Convert a value coming from a JSON blob to a target type """
        construct: typing.Callable[[Any], Any] | None = None

        # Extract type from "SomeType | None" unions
        if type(dstType) is UnionType:
            union = typing.get_args(dstType)
            assert len(union) == 2
            dstType = next(t for t in union if t is not NoneType)
        assert type(dstType) is not UnionType

        # Extract generic class from GenericAlias, e.g. list[str] --> list
        if type(dstType) is GenericAlias:
            dstType = typing.get_origin(dstType)
        assert not isinstance(dstType, GenericAlias)

        srcType: type
        if dstType is bytes:
            srcType = str
            construct = base64.b64decode
        elif dstType is set:
            srcType = list
        elif issubclass(dstType, enum.StrEnum):
            srcType = str
        elif issubclass(dstType, enum.IntEnum | enum.Enum):
            srcType = int
        elif dstType is Signature:
            srcType = dict
            construct = PrefsFile.decodeSignature
        else:
            srcType = dstType

        if srcType is not dstType:
            if not isinstance(o, srcType):
                raise ValueError("unexpected JSON field type")
            if construct is not None:
                o = construct(o)
            else:
                o = dstType(o)

        return o

    @staticmethod
    def decodeSignature(j: dict) -> Signature:
        name = str(j["name"])
        email = str(j["email"])
        time = int(j["time"])
        offset = int(j["offset"])
        return Signature(name, email, time, offset)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import os

from gitfourchette.gitdriver import GitStatus
from gitfourchette.graph import Graph, GraphWeaver
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import *
from gitfourchette.repomodel import RepoModel, UC_FAKEID
from gitfourchette.qt import *


class BlameModel:
    repoModel: RepoModel
    revList: RevList
    graph: Graph
    currentRevision: Revision
    revsFile: QTemporaryFile

    def __init__(self, repoModel: RepoModel, revList: RevList):
        self.repoModel = repoModel
        self.revList = revList

        # Create graph
        self.graph, graphWeaver = GraphWeaver.newGraph()
        for revision in revList.sequence:
            graphWeaver.newCommit(revision.commitId, revision.parentIds)
            self.graph.commitRows[revision.commitId] = graphWeaver.row

        # Fall back to newest revision
        self.currentRevision = revList.sequence[0]

        if APP_DEBUG:
            allCommitIds = {revision.commitId for revision in revList.sequence}
            assert len(allCommitIds) == len(revList.sequence), "duplicate commits in sequence"
            self.graph.testConsistency()

        # Dump revs file to speed up calls to 'git blame'
        # (and to make sure it won't return commits outside the trace)
        revsFileTemplate = os.path.join(qTempDir(), "blamerevs-XXXXXX.txt")
        self.revsFile = QTemporaryFile(revsFileTemplate, None)
        self.revsFile.open()
        self.revsFile.write(revList.serializeRevisionList().encode("utf-8"))
        self.revsFile.close()

    def prepareForDeletion(self):
        del self.repoModel

    @property
    def repo(self) -> Repo:
        return self.repoModel.repo


@dataclasses.dataclass
class Revision:
    """
    State of a file at a specific commit.
    Each line is annotated with the commit it originates from.
    """

    @dataclasses.dataclass(frozen=True)
    class BlameLine:
        commitId: Oid
        "Commit where this line appeared first."

        originalLineNumber: int
        "Line number in the original file in the source commit."

    path: str
    commitId: Oid
    parentIds: list[Oid] = dataclasses.field(default_factory=list)
    status: GitStatus = GitStatus.Modified
    blameLines: list[BlameLine] = dataclasses.field(default_factory=list)
    fullText: str | None = None
    binary: bool = False

    def toLocator(self) -> NavLocator:
        if self.commitId == UC_FAKEID:
            return NavLocator.inWorkdir(self.path)
        else:
            return NavLocator.inCommit(self.commitId, self.path)

    def isAnnotated(self) -> bool:
        return bool(self.blameLines)

    def findLine(self, target: BlameLine, start: int, searchRange: int = 250) -> int:
        assert self.isAnnotated()

        lines = self.blameLines
        count = len(lines)
        start = min(start, count - 1)
        searchRange = min(searchRange, count)

        lo = start
        hi = start + 1

        for _i in range(searchRange):
            if lo >= 0 and lines[lo] == target:
                return lo
            if hi < count and lines[hi] == target:
                return hi
            lo -= 1
            hi += 1

        raise ValueError("annotated line not found within given range")


class RevList:
    """
    History of significant revisions in a file, curated to only include commits
    that directly touched it.
    """

    sequence: list[Revision]
    byCommit: dict[Oid, Revision]
    nonTipCommits: set[Oid]

    def __init__(self):
        self.sequence = []
        self.byCommit = {}
        self.nonTipCommits = set()

    def insert(self, index: int, revision: Revision):
        self.sequence.insert(index, revision)
        self.byCommit[revision.commitId] = revision

    def push(self, revision: Revision):
        self.sequence.append(revision)
        self.byCommit[revision.commitId] = revision

    def __len__(self) -> int:
        return len(self.sequence)

    def revisionForCommit(self, oid: Oid) -> Revision:
        return self.byCommit[oid]

    def revisionNumber(self, oid: Oid) -> int:
        revision = self.byCommit[oid]
        index = self.sequence.index(revision)
        return len(self.sequence) - index

    def serializeRevisionList(self) -> str:
        """ Serialize the file's commit history (including parent rewriting)
        in a format suitable for "git blame -S <revs-file>". """
        lines = []
        for revision in self.sequence:
            parents = ' '.join(str(p) for p in revision.parentIds)
            lines.append(f"{revision.commitId} {parents}")
        return "\n".join(lines)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging

from gitfourchette import colors
from gitfourchette import settings
from gitfourchette.syntax import ColorScheme, LexJob
from gitfourchette.qt import *
from gitfourchette.toolbox import benchmark, CallbackAccumulator

logger = logging.getLogger(__name__)


class CodeHighlighter(QSyntaxHighlighter):
    scheme: ColorScheme
    lexJobs: list[LexJob]

    _whitespaceRegex = QRegularExpression("[\x09\x20\xA0]+")
    _whitespaceRegex.optimize()
    assert _whitespaceRegex.isValid()

    def __init__(self, parent):
        super().__init__(parent)

        self.scheme = ColorScheme()
        self.lexJobs = []

        self.occurrenceFormat = QTextCharFormat()
        self.occurrenceFormat.setBackground(colors.yellow)
        self.occurrenceFormat.setForeground(colors.black)
        self.occurrenceFormat.setFontWeight(QFont.Weight.Bold)

        self.searchTerm = ""
        self.numOccurrences = 0

    def setSearchTerm(self, term: str) -> int:
        if self.searchTerm != term:
            self.searchTerm = term
            self.numOccurrences = 0
            self.rehighlight()
        return self.numOccurrences

    def installLexJob(self, job):
        job.pulse.connect(self.onLexPulse)
        self.lexJobs.append(job)

    def stopLexJobs(self):
        for job in self.lexJobs:
            job.stop()
            job.pulse.disconnect(self.onLexPulse)
        self.lexJobs.clear()

    def highlightBlock(self, text: str):
        if self.scheme:
            # Highlight pygments tokens
            if self.lexJobs:
                self.highlightSyntax(text)

            # Override highlighting on whitespace with a nicer muted color
            if settings.prefs.showWhitespace:
                charFormat = self.scheme.whitespaceFormat
                wsIter = self._whitespaceRegex.globalMatch(text)
                while wsIter.hasNext():
                    match = wsIter.next()
                    self.setFormat(match.capturedStart(), match.capturedLength(), charFormat)

        # Override highlighting on search term
        if self.searchTerm:
            self.highlightSearch(text)

    def highlightSearch(self, text: str):
        # Highlight occurrences of search term
        term = self.searchTerm
        termLength = len(term)

        text = text.lower()
        textLength = len(text)

        index = 0
        while index < textLength:
            index = text.find(term, index)
            if index < 0:
                break
            self.setFormat(index, termLength, self.occurrenceFormat)
            index += termLength
            self.numOccurrences += 1

    def highlightSyntax(self, text: str):
        # Pygments syntax highlighting
        lineNumber = 1 + self.currentBlock().blockNumber()

        column = 0
        scheme = self.scheme.scheme
        lexJob = self.lexJobs[0]

        for tokenType, tokenLength in lexJob.tokens(lineNumber, text):
            try:
                charFormat = scheme[tokenType]
            except KeyError:
                charFormat = ColorScheme.fillInFallback(scheme, tokenType)
            self.setFormat(column, tokenLength, charFormat)
            column += tokenLength

    def setColorScheme(self, scheme: ColorScheme):
        self.scheme = scheme
        scheme.primeHighContrastVersion()

    @CallbackAccumulator.deferredMethod()
    @benchmark
    def onLexPulse(self):
        self.rehighlight()

    def onParentVisibilityChanged(self, visible: bool):
        """ Pause lexing when the parent DiffView is in the background """
        for job in self.lexJobs:
            if job.lexingComplete:
                pass
            elif visible:
                job.start()
            else:
                job.stop()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *
from gitfourchette.toolbox.iconbank import stockIcon


class QBusySpinner(QLabel):
    def __init__(self, parent):
        super().__init__(parent)
        self.setFixedSize(16, 16)
        self.currentFrame = 0
        self.numFrames = 8
        self.timer = QTimer(self)
        self.timer.setInterval(125)
        self.timer.timeout.connect(self.nextFrame)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

    def showEvent(self, event: QShowEvent):
        super().showEvent(event)
        if not self.timer.isActive():
            self.timer.start()
            self.nextFrame()

    def hideEvent(self, event: QHideEvent):
        self.timer.stop()
        super().hideEvent(event)

    def nextFrame(self):
        self.currentFrame = (self.currentFrame + 1) % self.numFrames
        self.updatePixmap()

    def updatePixmap(self):
        size = min(self.width(), self.height())
        icon = stockIcon(f"busyspinner{1 + self.currentFrame}")
        pixmap = icon.pixmap(QSize(size, size))
        self.setPixmap(pixmap)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from gitfourchette.qt import *
from gitfourchette.toolbox.iconbank import stockIcon

_IconFail = "achtung"
_IconPass = "input-validated"


class ValidatorMultiplexer(QObject):
    """
    Provides input validation in multiple QLineEdits and manages the enabled
    state of a set of so-called "gated widgets" depending on the validity of
    the inputs.

    Each QLineEdit is associated to a separate validator function that takes an
    input string (the contents of the QLineEdit) and returns an error string.
    If the validator function returns an empty error string, the QLineEdit is
    deemed to have valid input.

    When the input in a QLineEdit is invalid, all gated widgets are disabled.
    The user can continue entering text into the QLineEdit, but a warning icon
    will appear in it, along with a tooltip reporting the error returned by the
    validator function.

    The gated widgets are enabled only if all QLineEdits contain valid input.

    A typical use case of this class is to disable the OK button in a QDialog
    until all QLineEdits in it are valid.
    """

    CallbackFunc = Callable[[str], str]

    @dataclass
    class Input:
        widget: QLineEdit
        validate: ValidatorMultiplexer.CallbackFunc
        mustBeValid: bool
        indicator: QAction | None
        error: str = ""

    gatedWidgets: list[QWidget]
    inputs: list[Input]
    toolTipDelay: QTimer

    successText: str
    """ When the input is valid, show a green icon with this tooltip text.
    The green icon won't be shown if this string is empty. """

    def __init__(self, parent):
        super().__init__(parent)
        self.gatedWidgets = []
        self.inverseGatedWidgets = []
        self.inputs = []
        self.toolTipDelay = QTimer(self)
        self.toolTipDelay.setSingleShot(True)
        self.toolTipDelay.setInterval(500 if not APP_TESTMODE else 0)
        self.successText = ""

    def setGatedWidgets(self, *args: QWidget):
        self.gatedWidgets = list(args)

    def connectInput(
            self,
            edit: QLineEdit,
            validate: Callable[[str], str],
            showError: bool = True,
            mustBeValid: bool = True):
        assert isinstance(edit, QLineEdit)
        assert callable(validate)

        if showError:
            errorButton = edit.addAction(stockIcon(_IconFail), QLineEdit.ActionPosition.TrailingPosition)
            errorButton.setVisible(False)
            errorButton.setObjectName("ValidatorMultiplexerLineEditAction")

            self.toolTipDelay.timeout.connect(lambda: self.showToolTip(newInput, False))
            errorButton.triggered.connect(lambda _: self.showToolTip(newInput, True))
        else:
            errorButton = None

        newInput = ValidatorMultiplexer.Input(edit, validate, mustBeValid, errorButton)

        self.inputs.append(newInput)
        edit.textChanged.connect(self.run)

    def run(self, silenceEmptyWarnings=False):
        if any(i.indicator for i in self.inputs):
            self.toolTipDelay.stop()
            QToolTip.hideText()

        # Run validators on each input
        success = True
        for input in self.inputs:
            if not input.widget.isEnabled():  # Skip disabled inputs
                input.error = ""
                continue

            inputText = input.widget.text()
            input.error = input.validate(inputText)

            if input.error:
                success &= not input.mustBeValid
                if silenceEmptyWarnings and not inputText:
                    # Hide "cannot be empty" message, but do disable gated widgets if this input is required
                    input.error = ""

            if input.indicator:
                input.indicator.setIcon(stockIcon(_IconFail))
                input.indicator.setToolTip(input.error)
                input.indicator.setVisible(bool(input.error))

                # Schedule tooltip only if failed input has focus
                if input.error and input.widget.hasFocus():
                    self.toolTipDelay.start()

        # Enable/disable gated widgets depending on validation success
        for w in self.gatedWidgets:
            w.setEnabled(success)

        if success and self.successText:
            for input in self.inputs:
                input.indicator.setIcon(stockIcon(_IconPass))
                input.indicator.setToolTip(self.successText)
                input.indicator.setVisible(True)

    def showToolTip(self, input: ValidatorMultiplexer.Input, atMousePosition=True):
        self.toolTipDelay.stop()  # Prevent delayed warning from appearing after clicking errorButton

        if not input.error:
            return

        if atMousePosition:
            pos = QCursor.pos()
        else:
            pos = QPoint(0, input.widget.height() // 2)
            pos = input.widget.mapToGlobal(pos)

        # Don't pass a parent widget to QToolTip.showText otherwise tooltip vanishes too quickly
        QToolTip.showText(pos, input.error)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

from typing import ClassVar, Any

from gitfourchette import settings
from gitfourchette.codeview.codeview import CodeView
from gitfourchette.qt import *
from gitfourchette.syntax import LexJobCache, LexerCache, LexJob


class CodeWindow(QWidget):
    """
    Standalone window containing a vanilla CodeView.
    """

    _liveWindows: ClassVar[list[CodeWindow]] = []
    """Currently open CodeWindows. Wayland's quirks force us to use a None
    parent, so we must keep track of the window so it doesn't get garbage
    collected instantly."""

    def __init__(
            self,
            codeViewClass: type[CodeView] = CodeView,
            uniqueIdentifier: Any = None,
    ):
        # Don't parent the window to another widget, because Wayland forces the
        # child window to be on top of its parent at all times.
        super().__init__(None)

        self.setObjectName(self.__class__.__name__)
        self.setWindowFlag(Qt.WindowType.Window, True)

        # Keep reference around to avoid being GC'd instantly
        CodeWindow._liveWindows.append(self)

        codeView = codeViewClass(parent=self)
        codeView.setUpAsDetachedWindow()

        layout = QVBoxLayout(self)
        layout.setContentsMargins(QMargins())
        layout.setSpacing(0)
        layout.addWidget(codeView.searchBar)
        layout.addWidget(codeView)

        self.setWindowTitle(self.objectName())
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.resize(codeView.idealDetachedSize())

        self.codeView = codeView
        self.uniqueIdentifier = uniqueIdentifier

    def setPlainText(self, text: str, path: str):
        self.codeView.setPlainText(text)
        self.codeText = text

        lexJob = self._getLexJob(text, path)
        if lexJob is not None:
            self.codeView.highlighter.installLexJob(lexJob)
            self.codeView.highlighter.rehighlight()

    @classmethod
    def activateExistingWindow(cls, ident: Any) -> CodeWindow | None:
        try:
            window = next(w for w in cls._liveWindows if w.uniqueIdentifier == ident)
            window.activateWindow()
            return window
        except StopIteration:
            return None

    def _getLexJob(self, text: str, path: str) -> LexJob | None:
        if not settings.prefs.isSyntaxHighlightingEnabled():
            return None

        cacheKey = f"CodeWindow:{self.uniqueIdentifier}"

        try:
            return LexJobCache.get(cacheKey)
        except KeyError:
            pass

        lexer = LexerCache.getLexerFromPath(path, settings.prefs.pygmentsPlugins)
        if lexer is None:
            return None

        lexJob = LexJob(lexer, text, cacheKey)
        if lexJob is None:
            return None

        LexJobCache.put(lexJob)
        return lexJob

    def closeEvent(self, event: QCloseEvent):
        assert self in CodeWindow._liveWindows
        CodeWindow._liveWindows.remove(self)

        super().closeEvent(event)

        if QT5:  # Qt5 in test mode needs an extra push
            self.deleteLater()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import re
from itertools import zip_longest

from gitfourchette.graph import *
from gitfourchette.graph.graphbuilder import GraphBuildLoop, MockCommit, MockOid, Oid

PADDING = 2
ABRIDGMENT_THRESHOLD = 25


def padx(x):
    assert x >= 0
    return x * PADDING


class GraphDiagram:
    @staticmethod
    def parse(text: str):
        sequence, _heads = GraphDiagram.parseDefinition(text)
        return GraphBuildLoop().sendAll(sequence).graph

    @staticmethod
    def parseDefinition(text: str) -> tuple[list[MockCommit], set[Oid]]:
        sequence: list[MockCommit] = []
        seen: set[Oid] = set()
        heads: set[Oid] = set()
        commitsById: dict[Oid, MockCommit] = {}

        lines = re.split(r"\s+", text)

        for line in lines:
            line = line.strip()
            if not line:
                continue

            split = line.strip().split(":")
            assert 1 <= len(split) <= 2

            chainStr = split[0]
            assert chainStr
            assert "," not in chainStr

            try:
                assert "-" not in split[1]
                rootParents = MockOid.encodeAll(split[1].split(","))
            except IndexError:
                rootParents = []

            chain = MockOid.encodeAll(chainStr.split("-"))
            parents = [[c] for c in chain[1:]] + [rootParents]

            for oid, commitParents in zip(chain, parents, strict=True):
                assert all(c.id != oid for c in sequence), f"Commit hash appears twice in sequence! {oid}"
                mockCommit = MockCommit(oid, commitParents)
                sequence.append(mockCommit)
                if mockCommit.id not in seen:
                    heads.add(mockCommit.id)
                seen.update(mockCommit.parent_ids)
                commitsById[oid] = mockCommit

        for mockCommit in sequence:
            mockParents = [commitsById.get(p) for p in mockCommit.parent_ids]
            mockCommit.parents = mockParents  # type: ignore[attr-defined]

        return sequence, heads

    @staticmethod
    def diagram(graph: Graph, row0=0, maxRows=20, hiddenCommits=None, verbose=False):
        if not hiddenCommits:
            hiddenCommits = set()

        try:
            player = graph.startPlayback(row0)
        except StopIteration:
            return f"Won't draw graph because it's empty below row {row0}!"

        diagram = GraphDiagram()

        for _ in player:
            frame = player.sealCopy()
            if frame.commit in hiddenCommits:
                continue
            diagram.newFrame(frame, hiddenCommits, verbose)
            maxRows -= 1
            if maxRows < 0:
                break

        return diagram.bake()

    # -----------------------------------------------------------------

    def __init__(self):
        self.scanlines = []
        self.margins = []

    def reserve(self, x, y, fill=" "):
        assert len(fill) == 1
        for _ in range(len(self.scanlines), y + 1):
            self.scanlines.append([])
            self.margins.append([])
        scanline = self.scanlines[y]
        for _ in range(len(scanline), padx(x) + 1):
            scanline.append(fill)
        return scanline

    def plot(self, x, y, c):
        assert len(c) == 1
        scanline = self.reserve(x, y)
        assert scanline[padx(x)] not in "╭╮╰╯", "overwriting critical glyph!"
        scanline[padx(x)] = c

    def hline(self, x1, y, x2, fill="─"):
        assert len(fill) == 1
        left, right = min(x1, x2), max(x1, x2)
        scanline = self.reserve(right, y)
        for i in range(padx(left), padx(right) + 1):
            assert scanline[i] not in "╭╮╰╯", "overwriting critical glyph!"
            scanline[i] = fill

    def addMarginText(self, y, text):
        self.reserve(0, y)
        self.margins[y].append(text)

    def removeLastRow(self):
        self.scanlines.pop()
        self.margins.pop()

    def bake(self) -> str:
        if self.margins:
            numMargins = max(len(rowMargins) for rowMargins in self.margins)
        else:
            numMargins = 0
        marginWidths = [0] * numMargins
        for margins in self.margins:
            for i, mText in enumerate(margins):
                marginWidths[i] = max(marginWidths[i], len(mText))

        text = ""
        for margins, scanline in zip(self.margins, self.scanlines, strict=True):
            for mWidth, mText in zip_longest(reversed(marginWidths), reversed(margins)):
                mWidth, mText = mWidth or 0, mText or ""
                text += mText.rjust(mWidth) + " "
            text += ''.join(scanline).rstrip()
            text += "\n"
        text = text.removesuffix("\n")
        return text

    def newFrame(self, frame: Frame, hiddenCommits, verbose):
        upper = len(self.scanlines)
        lower = upper + 1
        homeLane = frame.homeLane()
        homeChain = frame.homeChain()

        for arc in frame.arcsPassingByCommit(hiddenCommits):
            # TODO: Depending on junctions, we may or may not want to abridge
            glyph = "│" if arc.length() <= ABRIDGMENT_THRESHOLD else "┊"
            self.plot(arc.lane, upper, glyph)
            self.plot(arc.lane, lower, glyph)

        closed = list(frame.arcsClosedByCommit(hiddenCommits))
        opened = list(frame.arcsOpenedByCommit(hiddenCommits))
        trivialClosed = True

        if closed:
            leftmostClosedLane = min([cl.lane for cl in closed])
            rightmostClosedLane = max([cl.lane for cl in closed])
            self.hline(leftmostClosedLane, upper, rightmostClosedLane)
            for cl in closed:
                if cl.lane != homeLane:
                    self.plot(cl.lane, upper, "╰╯"[cl.lane > homeLane])
                    trivialClosed = False

        if opened:
            row = upper if trivialClosed else lower
            leftmostOpenedLane = min([ol.lane for ol in opened])
            rightmostOpenedLane = max([ol.lane for ol in opened])
            self.hline(leftmostOpenedLane, row, rightmostOpenedLane)
            for ol in opened:
                if ol.lane == homeLane:
                    self.plot(ol.lane, row, "│")
                else:
                    self.plot(ol.lane, row, "╭╮"[ol.lane > homeLane])
                if row == upper:
                    self.plot(ol.lane, lower, "│")

        for arc, junction in frame.junctionsAtCommit(hiddenCommits):
            row = upper if trivialClosed else lower
            assert arc.junctions == sorted(arc.junctions), "junction list is supposed to be sorted!"
            assert junction.joinedBy == frame.commit, "junction commit != frame commit"
            assert homeLane != arc.lane, "junction plugged into passing arc that's on my homeLane?"
            self.hline(homeLane, row, arc.lane)
            self.plot(arc.lane, row, "╭╮"[homeLane < arc.lane])
            if row != upper:
                self.plot(homeLane, row, "╯╰"[homeLane < arc.lane])

        commitGlyph = "╳┷┯┿"[bool(opened) << 1 | bool(closed)]
        self.plot(homeLane, upper, commitGlyph)

        self.addMarginText(upper, str(frame.commit))
        if verbose:
            self.addMarginText(upper, str(int(frame.row)))
            self.addMarginText(upper, str(int(homeChain.topRow)))

        self.reserve(0, lower)  # make sure we're not removing upper row if we didn't plot anything in last row
        if not any(c in "╭╮╰╯" for c in self.scanlines[-1]):
            self.removeLastRow()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging

from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.codeview.codegutter import CodeGutter
from gitfourchette.codeview.codehighlighter import CodeHighlighter
from gitfourchette.codeview.coderubberband import CodeRubberBand
from gitfourchette.codeview.codesearch import CodeSearch
from gitfourchette.diffview.diffdocument import DiffTextFormats
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.qt import *
from gitfourchette.syntax import ColorScheme
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class CodeView(QPlainTextEdit):
    contextualHelp = Signal(str)
    selectionActionable = Signal(bool)
    visibilityChanged = Signal(bool)
    sizeChanged = Signal()

    highlighter: CodeHighlighter
    gutter: CodeGutter
    currentLocator: NavLocator
    isDetachedWindow: bool

    FormattingMarkFlags = QTextOption.Flag.ShowTabsAndSpaces

    def __init__(
            self,
            gutterClass: type[CodeGutter] = CodeGutter,
            highlighterClass: type[CodeHighlighter] = CodeHighlighter,
            parent=None
    ):
        super().__init__(parent)

        self.setReadOnly(True)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.TextSelectableByKeyboard)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onContextMenuRequested)

        self.currentLocator = NavLocator()
        self.isDetachedWindow = False

        # Highlighter for search terms
        self.highlighter = highlighterClass(self)
        self.highlighter.setDocument(self.document())
        self.visibilityChanged.connect(self.highlighter.onParentVisibilityChanged)

        self.gutter = gutterClass(self)
        self.gutter.customContextMenuRequested.connect(self.onContextMenuRequestedFromGutter)
        self.updateRequest.connect(self.gutter.onParentUpdateRequest)
        # self.blockCountChanged.connect(self.updateGutterWidth)
        self.syncViewportMarginsWithGutter()

        self.cursorPositionChanged.connect(self.updateRubberBand)
        self.selectionChanged.connect(self.updateRubberBand)

        searchProvider = CodeSearch(self)

        self.searchBar = SearchBar(self, searchProvider)
        self.searchBar.hide()

        self.rubberBand = CodeRubberBand(self.viewport())
        self.rubberBand.hide()

        self.rubberBandButtonGroup = QWidget(parent=self.viewport())
        rubberBandButtonLayout = QHBoxLayout(self.rubberBandButtonGroup)
        rubberBandButtonLayout.setSpacing(0)
        rubberBandButtonLayout.setContentsMargins(0, 0, 0, 0)
        self.rubberBandButtonGroup.hide()

        self.verticalScrollBar().valueChanged.connect(self.updateRubberBand)
        self.horizontalScrollBar().valueChanged.connect(self.updateRubberBand)

        # Initialize font & styling
        GFApplication.instance().restyle.connect(self.refreshPrefs)
        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

        makeWidgetShortcut(self, self.onEscapeKey, "Escape")

    def setUpAsDetachedWindow(self):
        # In a detached window, we can't rely on the main window's menu bar to
        # dispatch shortcuts to us (except on macOS, which has a global main menu).

        self.isDetachedWindow = True

        bar = self.searchBar
        makeWidgetShortcut(self, lambda: bar.popUp(SearchBar.Op.Start), *GlobalShortcuts.find)
        makeWidgetShortcut(self, lambda: bar.popUp(SearchBar.Op.Next), *GlobalShortcuts.findNext)
        makeWidgetShortcut(self, lambda: bar.popUp(SearchBar.Op.Previous), *GlobalShortcuts.findPrevious)

        makeWidgetShortcut(self, lambda: self.window().close(), QKeySequence.StandardKey.Close,
                           context=Qt.ShortcutContext.WindowShortcut)

    # ---------------------------------------------
    # Qt events

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        self.resizeGutter()
        self.updateRubberBand()
        self.sizeChanged.emit()

    def wheelEvent(self, event: QWheelEvent):
        # Drop-in replacement for QPlainTextEdit::wheelEvent which scales text
        # on ctrl+wheel. The vanilla version doesn't emit a signal, but we need
        # to percolate the new font to the gutter & rubberband.
        # See https://github.com/qt/qtbase/blob/6.7.2/src/widgets/widgets/qplaintextedit.cpp#L2327
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            self.wheelZoom(event.angleDelta().y())
            return

        super().wheelEvent(event)

    def wheelZoom(self, delta: int):
        delta //= 120
        font = self.font()
        oldSize = font.pointSizeF()
        newSize = oldSize + delta
        newSize = max(4, newSize)
        if newSize == oldSize:
            return
        font.setPointSizeF(newSize)
        self.setFont(font)
        self.gutter.syncFont(font)
        self.syncViewportMarginsWithGutter()
        self.updateRubberBand()

    def focusInEvent(self, event: QFocusEvent):
        self.rubberBand.repaint()
        super().focusInEvent(event)

    def focusOutEvent(self, event: QFocusEvent):
        self.rubberBand.repaint()
        super().focusOutEvent(event)

    def showEvent(self, event: QShowEvent):
        super().showEvent(event)
        self.visibilityChanged.emit(True)

    def hideEvent(self, event: QHideEvent):
        super().hideEvent(event)
        self.visibilityChanged.emit(False)

    def onEscapeKey(self):
        if self.isDetachedWindow and not self.searchBar.isVisible():
            self.window().close()
        else:
            self.searchBar.hideOrBeep()

    # ---------------------------------------------
    # Document replacement

    def clear(self):  # override
        # Clear info about the current patch - necessary for document reuse detection to be correct when the user
        # clears the selection in a FileList and then reselects the last-displayed document.
        self.currentLocator = NavLocator()

        # Clear the actual contents
        super().clear()

    # ---------------------------------------------
    # Restore position

    def restorePosition(self, locator: NavLocator):
        # Get position at start/end of line
        block: QTextBlock = self.document().findBlockByNumber(locator.cursorLine)
        sol = block.position()
        eol = block.position() + block.length()

        # If cursor position still falls within the same line, keep that position.
        # Otherwise, snap cursor position to start of line.
        pos = locator.cursorChar
        if not (sol <= pos < eol):
            pos = sol

        # Move text cursor
        newTextCursor = self.textCursor()
        newTextCursor.setPosition(pos)
        self.setTextCursor(newTextCursor)

        # Restore the scrollbar
        self.restoreScrollPosition(locator.scrollChar)

    def restoreScrollPosition(self, topCharacter: int) -> int:
        scrollValue = self._findScrollPosition(topCharacter)
        self.verticalScrollBar().setValue(scrollValue)
        return scrollValue

    def _findScrollPosition(self, topCharacter: int) -> int:
        """
        Return a value to be set on the vertical scrollbar so that `topCharacter`
        is part of the first visible line in the viewport.
        """
        if topCharacter == 0:
            return 0

        topCursor = self.textCursor()
        topCursor.setPosition(topCharacter)

        scrollBar = self.verticalScrollBar()

        # If line wrapping is OFF, QScrollBar.value() perfectly matches up with
        # line numbers.
        if self.lineWrapMode() == QPlainTextEdit.LineWrapMode.NoWrap:
            return topCursor.blockNumber()

        # If line wrapping is ON, we can't simply cache and restore QScrollBar.value().
        # It seems that QPlainTextEdit doesn't wrap lines that aren't visible yet,
        # so the scrollbar's range isn't reliable until the desired line is in view.

        # First, center the viewport on the desired top character.
        backupCursor = self.textCursor()
        self.setTextCursor(topCursor)
        self.centerCursor()

        # Scroll down one notch at a time until topCharacter becomes part of
        # the top visible line in the viewport.
        cornerPixel = self.topLeftCornerPixel()
        scrollValue = 0
        i = 0
        while i < 500 and self.cursorForPosition(cornerPixel).position() < topCharacter:
            i += 1
            scrollValue = scrollBar.value() + 1
            scrollBar.setValue(scrollValue)
            if scrollBar.value() < scrollValue:  # Can't scroll past end of document
                break

        # logger.debug(f"Stabilized in {i} iterations - final scroll {scrollValue} - "
        #              f"char pos {self.cursorForPosition(cornerPixel).position()} vs {topCharacter}")

        # Restore backup cursor
        self.setTextCursor(backupCursor)

        return scrollValue

    def topLeftCornerPixel(self) -> QPoint:
        return QPoint(0, self.fontMetrics().height() // 2)

    def topLeftCornerCursor(self) -> QTextCursor:
        cornerPixel = self.topLeftCornerPixel()
        cornerCursor = self.cursorForPosition(cornerPixel)
        return cornerCursor

    def topLeftCornerCharacter(self) -> int:
        cornerCursor = self.topLeftCornerCursor()
        return cornerCursor.position()

    def preciseLocator(self) -> NavLocator:
        scrollChar = self.topLeftCornerCharacter()
        textCursor = self.textCursor()

        locator = self.currentLocator.coarse()
        locator = locator.replace(
            cursorChar=textCursor.position(),
            cursorLine=textCursor.blockNumber(),
            scrollChar=scrollChar)

        return locator

    # ---------------------------------------------
    # Prefs

    def refreshPrefs(self, changeColorScheme=True):
        monoFont = settings.prefs.monoFont()
        self.setFont(monoFont)

        currentDocument = self.document()
        if currentDocument:
            currentDocument.setDefaultFont(monoFont)

        tabWidth = settings.prefs.tabSpaces
        self.setTabStopDistance(QFontMetricsF(monoFont).horizontalAdvance(' ' * tabWidth))
        self.refreshWordWrap()
        self.refreshShowWhitespace()
        self.setCursorWidth(2)

        self.gutter.syncFont(monoFont)
        self.syncViewportMarginsWithGutter()

        if changeColorScheme:
            scheme = settings.prefs.syntaxHighlightingScheme()
            self.setColorScheme(scheme)

    def setColorScheme(self, scheme: ColorScheme):
        DiffTextFormats.refresh(scheme, settings.prefs.colorblind)

        self.highlighter.setColorScheme(scheme)
        self.highlighter.rehighlight()

        # See selection-background-color in .qss asset.
        dark = scheme.isDark() if scheme else isDarkTheme()
        self.setProperty("dark", "true" if dark else "false")

        # Had better luck setting colors with a stylesheet than via setPalette().
        styleSheet = scheme.basicQss(self)
        self.setStyleSheet(styleSheet)

    def refreshWordWrap(self):
        if settings.prefs.wordWrap:
            wrapMode = QPlainTextEdit.LineWrapMode.WidgetWidth
        else:
            wrapMode = QPlainTextEdit.LineWrapMode.NoWrap

        # Bail if no-op
        if self.lineWrapMode() == wrapMode:
            return

        # Changing the wrap mode will trash our scroll position, so remember where we are.
        topCharacter = self.topLeftCornerCharacter()

        self.setLineWrapMode(wrapMode)
        self.restoreScrollPosition(topCharacter)

    def refreshShowWhitespace(self):
        doc = self.document()
        if doc is None:
            return
        opt = QTextOption(doc.defaultTextOption())
        flags = opt.flags()
        if settings.prefs.showWhitespace:
            flags |= self.FormattingMarkFlags
        else:
            flags &= ~self.FormattingMarkFlags
        opt.setFlags(flags)
        doc.setDefaultTextOption(opt)

    # ---------------------------------------------
    # Context menu

    def contextMenuActions(self, clickedCursor: QTextCursor) -> list[ActionDef]:
        if type(self) is not CodeView:  # pragma: no cover
            raise NotImplementedError("CodeView subclasses must override this function")

        return []

    def onContextMenuRequested(self, point: QPoint):
        # Don't show the context menu if we're empty
        if self.document().isEmpty():
            return

        # Get standard context menu (copy, select all, etc.)
        bottomMenu: QMenu = self.createStandardContextMenu()

        # Get position of click in document
        clickedCursor = self.cursorForPosition(point)

        # Get actions from concrete class
        specializedActions = self.contextMenuActions(clickedCursor)

        # Append common CodeView actions
        actions: list[ActionDef | QAction] = [
            *specializedActions,
            ActionDef.SEPARATOR,
            *[a for a in bottomMenu.actions() if not a.isSeparator()],
            ActionDef.SEPARATOR,
            ActionDef(_("Configure Appearance…"), lambda: GFApplication.instance().openPrefsDialog("font"), icon="configure"),
        ]

        # Create QMenu
        menu = ActionDef.makeQMenu(self, actions)
        menu.setObjectName("CodeViewContextMenu")
        menu.aboutToHide.connect(menu.deleteLater)

        # The master menu doesn't take ownership of bottomMenu's actions,
        # so bottomMenu must be kept alive until the master menu is closed.
        menu.destroyed.connect(bottomMenu.deleteLater)

        # Show QMenu
        menu.popup(self.viewport().mapToGlobal(point))

    def onContextMenuRequestedFromGutter(self, point: QPoint):
        point = self.gutter.mapToGlobal(point)
        point = self.viewport().mapFromGlobal(point)
        self.onContextMenuRequested(point)

    # ---------------------------------------------
    # Gutter

    def resizeGutter(self):
        cr: QRect = self.contentsRect()
        cr.setWidth(self.gutter.calcWidth())
        self.gutter.setGeometry(cr)

    def syncViewportMarginsWithGutter(self):
        self.gutter.refreshMetrics()
        gutterWidth = self.gutter.calcWidth()

        # Prevent Qt freeze if margin width exceeds widget width, e.g. when window is very narrow
        # (especially prevalent with word wrap?)
        self.setMinimumWidth(gutterWidth * 2)

        self.setViewportMargins(gutterWidth, 0, 0, 0)

    # ---------------------------------------------
    # Rubberband

    def updateRubberBand(self):
        pass

    # ---------------------------------------------
    # Cursor/selection

    def getSelectedLineExtents(self) -> tuple[int, int]:
        """ Return block numbers of the first and last blocks encompassing the current selection """

        cursor: QTextCursor = self.textCursor()
        posStart = cursor.selectionStart()
        posEnd = cursor.selectionEnd()

        assert posStart >= 0
        assert posEnd >= 0

        document: QTextDocument = self.document()
        startBlock = document.findBlock(posStart).blockNumber()
        endBlock = document.findBlock(posEnd).blockNumber()

        return startBlock, endBlock

    def getMaxPosition(self) -> int:
        lastBlock = self.document().lastBlock()
        return lastBlock.position() + max(0, lastBlock.length() - 1)

    def getAnchorHomeLinePosition(self) -> int:
        cursor: QTextCursor = self.textCursor()

        # Snap anchor to start of home line
        cursor.setPosition(cursor.anchor(), QTextCursor.MoveMode.MoveAnchor)
        cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock, QTextCursor.MoveMode.MoveAnchor)

        return cursor.anchor()

    def getStartOfLineAt(self, point: QPoint) -> int:
        clickedCursor: QTextCursor = self.cursorForPosition(point)
        clickedCursor.movePosition(QTextCursor.MoveOperation.StartOfBlock)
        return clickedCursor.position()

    def replaceCursor(self, cursor: QTextCursor):
        """Replace the cursor without moving the horizontal scroll bar"""
        with QScrollBackupContext(self.horizontalScrollBar()):
            self.setTextCursor(cursor)

    def selectWholeLineAt(self, point: QPoint):
        clickedPosition = self.getStartOfLineAt(point)

        cursor: QTextCursor = self.textCursor()
        cursor.setPosition(clickedPosition)
        cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock, QTextCursor.MoveMode.KeepAnchor)

        self.replaceCursor(cursor)

    def selectWholeLinesTo(self, point: QPoint):
        homeLinePosition = self.getAnchorHomeLinePosition()
        clickedPosition = self.getStartOfLineAt(point)

        cursor: QTextCursor = self.textCursor()

        if homeLinePosition <= clickedPosition:
            # Move anchor to START of home line
            cursor.setPosition(homeLinePosition, QTextCursor.MoveMode.MoveAnchor)
            # Move cursor to END of clicked line
            cursor.setPosition(clickedPosition, QTextCursor.MoveMode.KeepAnchor)
            cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock, QTextCursor.MoveMode.KeepAnchor)
        else:
            # Move anchor to END of home line
            cursor.setPosition(homeLinePosition, QTextCursor.MoveMode.MoveAnchor)
            cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock, QTextCursor.MoveMode.MoveAnchor)
            # Move cursor to START of clicked line
            cursor.setPosition(clickedPosition, QTextCursor.MoveMode.KeepAnchor)
            cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock, QTextCursor.MoveMode.KeepAnchor)

        self.replaceCursor(cursor)

    @staticmethod
    def currentDetachedCodeView() -> CodeView:
        activeWindow = QApplication.activeWindow()
        if not activeWindow:
            raise KeyError("no active window")
        detachedCodeView: CodeView = activeWindow.findChild(CodeView)
        if detachedCodeView is None:
            raise KeyError("no detached code view")
        assert detachedCodeView.isDetachedWindow
        return detachedCodeView

    def idealDetachedSize(self) -> QSize:
        windowHeight = int(QApplication.primaryScreen().availableSize().height() * .8)
        windowWidth = (self.gutter.calcWidth()
                       + self.fontMetrics().horizontalAdvance("M" * 81)
                       + self.verticalScrollBar().width())
        return QSize(windowWidth, windowHeight)




# Form implementation generated from reading ui file 'resetheaddialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_ResetHeadDialog(object):
    def setupUi(self, ResetHeadDialog):
        ResetHeadDialog.setObjectName("ResetHeadDialog")
        ResetHeadDialog.resize(600, 325)
        self.gridLayout = QGridLayout(ResetHeadDialog)
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setVerticalSpacing(20)
        self.gridLayout.setObjectName("gridLayout")
        self.softButton = QRadioButton(parent=ResetHeadDialog)
        self.softButton.setAutoFillBackground(False)
        self.softButton.setText("&Soft")
        self.softButton.setObjectName("softButton")
        self.gridLayout.addWidget(self.softButton, 0, 0, 1, 1, Qt.AlignmentFlag.AlignTop)
        self.softHelp = QLabel(parent=ResetHeadDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.softHelp.sizePolicy().hasHeightForWidth())
        self.softHelp.setSizePolicy(sizePolicy)
        self.softHelp.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.softHelp.setWordWrap(True)
        self.softHelp.setObjectName("softHelp")
        self.gridLayout.addWidget(self.softHelp, 0, 1, 1, 1)
        self.buttonBox = QDialogButtonBox(parent=ResetHeadDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 5, 0, 1, 2)
        self.hardButton = QRadioButton(parent=ResetHeadDialog)
        self.hardButton.setText("&Hard")
        self.hardButton.setObjectName("hardButton")
        self.gridLayout.addWidget(self.hardButton, 2, 0, 1, 1, Qt.AlignmentFlag.AlignTop)
        self.mixedHelp = QLabel(parent=ResetHeadDialog)
        self.mixedHelp.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.mixedHelp.setWordWrap(True)
        self.mixedHelp.setObjectName("mixedHelp")
        self.gridLayout.addWidget(self.mixedHelp, 1, 1, 1, 1)
        self.hardHelp = QLabel(parent=ResetHeadDialog)
        self.hardHelp.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.hardHelp.setWordWrap(True)
        self.hardHelp.setObjectName("hardHelp")
        self.gridLayout.addWidget(self.hardHelp, 2, 1, 1, 1)
        self.mixedButton = QRadioButton(parent=ResetHeadDialog)
        self.mixedButton.setText("&Mixed")
        self.mixedButton.setObjectName("mixedButton")
        self.gridLayout.addWidget(self.mixedButton, 1, 0, 1, 1, Qt.AlignmentFlag.AlignTop)
        self.recurseCheckBox = QCheckBox(parent=ResetHeadDialog)
        self.recurseCheckBox.setChecked(True)
        self.recurseCheckBox.setObjectName("recurseCheckBox")
        self.gridLayout.addWidget(self.recurseCheckBox, 3, 0, 1, 2)

        self.retranslateUi(ResetHeadDialog)
        self.buttonBox.accepted.connect(ResetHeadDialog.accept)
        self.buttonBox.rejected.connect(ResetHeadDialog.reject)
        ResetHeadDialog.setTabOrder(self.softButton, self.mixedButton)
        ResetHeadDialog.setTabOrder(self.mixedButton, self.hardButton)

    def retranslateUi(self, ResetHeadDialog):
        ResetHeadDialog.setWindowTitle(_p("ResetHeadDialog", "Reset HEAD"))
        self.softHelp.setText(_p("ResetHeadDialog", "<b>Keep</b> all your files intact.\n"
"<br><b>Stage</b> any differences between your files and {commit}."))
        self.mixedHelp.setText(_p("ResetHeadDialog", "<b>Keep</b> all your files intact.\n"
"<br><b>Unstage</b> any differences between your files and {commit}."))
        self.hardHelp.setText(_p("ResetHeadDialog", "<b>Wipe</b> all your changes.\n"
"<br>End result identical to a fresh checkout of {commit}."))
        self.recurseCheckBox.setText(_p("ResetHeadDialog", "&Reset submodules recursively"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging

from gitfourchette.forms.newbranchdialog import NewBranchDialog
from gitfourchette.forms.resetheaddialog import ResetHeadDialog
from gitfourchette.forms.textinputdialog import TextInputDialog
from gitfourchette.gitdriver import argsIf
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskPrereqs, TaskEffects
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class SwitchBranch(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoConflicts

    def flow(
            self,
            newBranch: str,
            askForConfirmation: bool = True,
            recurseSubmodules: bool = False,
            refreshUnderDetachedWarning: bool = False):
        assert not newBranch.startswith(RefPrefix.HEADS)

        branchObj: Branch = self.repo.branches.local[newBranch]

        if branchObj.is_checked_out():
            message = _("Branch {0} is already checked out.", bquo(newBranch))
            raise AbortTask(message, 'information')

        if askForConfirmation:
            text = _("Do you want to switch to branch {0}?", bquo(newBranch))
            verb = _("Switch")

            recurseCheckbox = None
            anySubmodules = bool(self.repo.listall_submodules_fast())
            if anySubmodules:
                recurseCheckbox = QCheckBox(_("Update submodules recursively"))
                recurseCheckbox.setChecked(True)

            yield from self.flowConfirm(text=text, verb=verb, checkbox=recurseCheckbox)
            recurseSubmodules = recurseCheckbox is not None and recurseCheckbox.isChecked()

        headId = self.repoModel.headCommitId
        if self.repoModel.dangerouslyDetachedHead() and branchObj.target != headId:
            if refreshUnderDetachedWarning:  # Refresh GraphView underneath dialog
                from gitfourchette.tasks import RefreshRepo
                yield from self.flowSubtask(RefreshRepo)

            text = paragraphs(
                _("You are in <b>Detached HEAD</b> mode at commit {0}.", btag(shortHash(headId))),
                _("You might lose track of this commit if you switch to {0}.", hquo(newBranch)))
            yesText = _("Switch to {0}", lquoe(newBranch))
            noText = _("Don’t Switch")
            yield from self.flowConfirm(text=text, icon='warning', verb=yesText, cancelText=noText)

        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Head

        yield from self.flowCallGit(
            "checkout",
            "--progress",
            "--no-guess",
            *argsIf(recurseSubmodules, "--recurse-submodules"),
            newBranch)

        self.epilog.status = _("Switched to branch {0}.", tquo(newBranch))


class RenameBranch(RepoTask):
    def flow(self, oldBranchName: str):
        assert not oldBranchName.startswith(RefPrefix.HEADS)

        forbiddenBranchNames = self.repo.listall_branches(BranchType.LOCAL)
        forbiddenBranchNames.remove(oldBranchName)

        nameTaken = _("This name is already taken by another local branch.")

        dlg = TextInputDialog(
            self.parentWidget(),
            _("Rename local branch"),
            _("Enter new name:"),
            subtitle=_("Current name: {0}", oldBranchName))
        dlg.setText(oldBranchName)
        dlg.setValidator(lambda name: nameValidationMessage(name, forbiddenBranchNames, nameTaken))
        dlg.lineEdit.setValidator(ReplaceSpacesWithDashes())
        dlg.okButton.setText(_("Rename"))

        # Pre-select leaf name (e.g. in "folder/leaf" select only "leaf")
        NewBranchDialog.preSelectLeaf(dlg.lineEdit)

        yield from self.flowDialog(dlg)
        dlg.deleteLater()
        newBranchName = dlg.lineEdit.text()

        # Bail if identical to dodge AlreadyExistsError
        if newBranchName == oldBranchName:
            raise AbortTask()

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Refs

        self.repo.rename_local_branch(oldBranchName, newBranchName)

        self.epilog.status = _("Branch {0} renamed to {1}.", tquo(oldBranchName), tquo(newBranchName))


class RenameBranchFolder(RepoTask):
    def flow(self, oldFolderRefName: str):
        prefix, oldFolderName = RefPrefix.split(oldFolderRefName)
        assert prefix == RefPrefix.HEADS
        assert not oldFolderName.endswith("/")
        oldFolderNameSlash = oldFolderName + "/"

        forbiddenBranches = set()
        folderBranches = []
        for oldBranchName in self.repo.listall_branches(BranchType.LOCAL):
            if oldBranchName.startswith(oldFolderNameSlash):
                folderBranches.append(oldBranchName)
            else:
                forbiddenBranches.add(oldBranchName)

        def transformBranchName(branchName: str, newFolderName: str) -> str:
            assert branchName.startswith(oldFolderName)
            newBranchName = newFolderName + branchName.removeprefix(oldFolderName)
            newBranchName = newBranchName.removeprefix("/")
            return newBranchName

        def validate(newFolderName: str) -> str:
            for oldBranchName in folderBranches:
                newBranchName = transformBranchName(oldBranchName, newFolderName)
                if newBranchName in forbiddenBranches:
                    return _("This name clashes with existing branch {0}.", tquo(newBranchName))
            # Finally validate the folder name itself as if it were a branch,
            # but don't test against existing refs (which we just did above),
            # and allow an empty name.
            if not newFolderName:
                return ""
            return nameValidationMessage(newFolderName, [])

        subtitle = _n("Folder {name} contains {n} branch.",
                      "Folder {name} contains {n} branches.",
                      len(folderBranches), name=lquoe(oldFolderName))

        dlg = TextInputDialog(
            self.parentWidget(),
            _("Rename branch folder"),
            _("Enter new name:"),
            subtitle=subtitle)
        dlg.setText(oldFolderName)
        dlg.setValidator(validate)
        dlg.okButton.setText(_("Rename"))
        dlg.lineEdit.setPlaceholderText(_("Leave blank to move the branches to the root folder."))
        dlg.lineEdit.setValidator(ReplaceSpacesWithDashes())

        yield from self.flowDialog(dlg)
        dlg.deleteLater()

        newFolderName = dlg.lineEdit.text()

        # Bail if identical to dodge AlreadyExistsError
        if newFolderName == oldFolderName:
            raise AbortTask()

        # Perform rename
        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Refs

        for oldBranchName in folderBranches:
            newBranchName = transformBranchName(oldBranchName, newFolderName)
            self.repo.rename_local_branch(oldBranchName, newBranchName)

        self.epilog.status = (
                _("Branch folder {0} renamed to {1}.", tquo(oldFolderName), tquo(newFolderName))
                + " "
                + _n("{n} branch affected.", "{n} branches affected.", len(folderBranches))
        )


class DeleteBranch(RepoTask):
    def flow(self, localBranchName: str):
        assert not localBranchName.startswith(RefPrefix.HEADS)

        if localBranchName == self.repo.head_branch_shorthand:
            text = paragraphs(
                _("Cannot delete {0} because it is the current branch.", bquo(localBranchName)),
                _("Before you try again, switch to another branch."))
            raise AbortTask(text)

        text = paragraphs(_("Really delete local branch {0}?", bquo(localBranchName)),
                          _("This cannot be undone!"))

        yield from self.flowConfirm(
            text=text,
            verb=_("Delete branch"),
            buttonIcon="SP_DialogDiscardButton")

        yield from self.flowEnterWorkerThread()
        target = self.repo.branches[localBranchName].target
        self.epilog.effects |= TaskEffects.Refs
        self.repo.delete_local_branch(localBranchName)

        self.epilog.status = _("Branch {0} deleted (commit at tip was {1}).",
                               tquo(localBranchName), tquo(shortHash(target)))


class DeleteBranchFolder(RepoTask):
    def flow(self, folderRefName: str):
        prefix, folderName = RefPrefix.split(folderRefName)
        assert prefix == RefPrefix.HEADS
        assert not folderName.endswith("/")
        folderNameSlash = folderName + "/"

        currentBranch = self.repo.head_branch_shorthand
        if currentBranch.startswith(folderNameSlash):
            text = paragraphs(
                _("Cannot delete folder {0} because it contains the current branch {1}.", bquo(folderName), bquo(currentBranch)),
                _("Before you try again, switch to another branch."))
            raise AbortTask(text)

        folderBranches = [b for b in self.repo.listall_branches(BranchType.LOCAL)
                          if b.startswith(folderNameSlash)]

        text = paragraphs(
            _("Really delete local branch folder {0}?", bquo(folderName)),
            _n("<b>{n}</b> branch will be deleted.", "<b>{n}</b> branches will be deleted.", n=len(folderBranches)
               ) + " " + _("This cannot be undone!"))

        yield from self.flowConfirm(
            _("Delete branch folder"),
            text,
            detailList=folderBranches,
            verb=_("Delete folder"),
            buttonIcon="SP_DialogDiscardButton")

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Refs

        for b in folderBranches:
            self.repo.delete_local_branch(b)

        self.epilog.status = _n("{n} branch deleted in folder {name}.",
                             "{n} branches deleted in folder {name}.",
                                n=len(folderBranches), name=tquo(folderName))


class NewBranchFromCommit(RepoTask):
    def flow(self, tip: Oid, localName: str = "", suggestUpstream: str = "", checkUpstream: bool = False):
        repo = self.repo

        tipHashText = shortHash(tip)

        # Are we creating a branch at the tip of the current branch?
        if not repo.head_is_unborn and not repo.head_is_detached and repo.head.target == tip:
            # Let user know that's the HEAD
            tipHashText = f"HEAD ({tipHashText})"

            # Default to the current branch's name (if no name given)
            if not localName:
                localName = repo.head.shorthand

        # Collect upstream names and set initial localName (if we haven't been able to set it above).
        refsPointingHere = repo.listall_refs_pointing_at(tip)
        upstreams = []
        for r in refsPointingHere:
            prefix, shorthand = RefPrefix.split(r)
            if prefix == RefPrefix.HEADS:
                if not localName:
                    localName = shorthand
                branch = repo.branches[shorthand]
                if branch.upstream:
                    upstreams.append(branch.upstream.shorthand)
            elif prefix == RefPrefix.REMOTES:
                if not localName:
                    _prefix, localName = split_remote_branch_shorthand(shorthand)
                upstreams.append(shorthand)

        # Start with a unique name so the branch validator doesn't shout at us
        forbiddenBranchNames = repo.listall_branches(BranchType.LOCAL)
        localName = withUniqueSuffix(localName, forbiddenBranchNames)

        # Ensure no duplicate upstreams (stable order since Python 3.7+)
        upstreams = list(dict.fromkeys(upstreams))

        forbiddenBranchNames = repo.listall_branches(BranchType.LOCAL)

        commitMessage = repo.get_commit_message(tip)
        commitMessage, _junk = messageSummary(commitMessage)

        showSubmoduleControls = bool(repo.listall_submodules_fast())

        dlg = NewBranchDialog(
            initialName=localName,
            target=tipHashText,
            targetSubtitle=commitMessage,
            upstreams=upstreams,
            reservedNames=forbiddenBranchNames,
            allowSwitching=not self.repo.any_conflicts,
            showSubmoduleControls=showSubmoduleControls,
            parent=self.parentWidget())

        if suggestUpstream:
            upstreamIndex = dlg.ui.upstreamComboBox.findText(suggestUpstream)
            if upstreamIndex >= 0:
                dlg.ui.upstreamCheckBox.setChecked(checkUpstream)
                dlg.ui.upstreamComboBox.setCurrentIndex(upstreamIndex)

        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        dlg.show()
        yield from self.flowDialog(dlg)
        dlg.deleteLater()

        localName = dlg.ui.nameEdit.text()
        switchTo = dlg.ui.switchToBranchCheckBox.isChecked()
        recurseSubmodules = dlg.ui.recurseSubmodulesCheckBox.isChecked()
        trackUpstream = ""
        if dlg.ui.upstreamCheckBox.isChecked():
            trackUpstream = dlg.ui.upstreamComboBox.currentText()

        yield from self.flowEnterWorkerThread()

        # Create local branch
        repo.create_branch_from_commit(localName, tip)
        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Upstreams
        self.epilog.status = _("Branch {0} created on commit {1}.", tquo(localName), tquo(shortHash(tip)))

        # Optionally make it track a remote branch
        if trackUpstream:
            repo.edit_upstream_branch(localName, trackUpstream)

        # Switch to it last (if user wants to)
        if switchTo:
            self.epilog.effects |= TaskEffects.Head
            yield from self.flowEnterUiThread()
            yield from self.flowSubtask(
                SwitchBranch,
                localName,
                askForConfirmation=False,
                recurseSubmodules=recurseSubmodules,
                refreshUnderDetachedWarning=True)


class NewBranchFromHead(RepoTask):
    def prereqs(self):
        return TaskPrereqs.NoUnborn

    def flow(self):
        if self.repo.head_is_detached:
            yield from self.flowSubtask(NewBranchFromCommit, self.repo.head_commit_id)
        else:
            yield from self.flowSubtask(NewBranchFromRef, self.repo.head_branch_fullname)


class NewBranchFromRef(RepoTask):
    def flow(self, refname: str):
        prefix, name = RefPrefix.split(refname)

        if prefix == RefPrefix.HEADS:
            branch = self.repo.branches.local[name]
            upstream = branch.upstream.shorthand if branch.upstream else ""
            tickUpstream = False

        elif prefix == RefPrefix.REMOTES:
            branch = self.repo.branches.remote[name]
            upstream = branch.shorthand
            name = name.removeprefix(branch.remote_name + "/")
            tickUpstream = True

        else:
            raise NotImplementedError(f"Unsupported prefix for refname '{refname}'")

        yield from self.flowSubtask(NewBranchFromCommit, branch.target, name,
                                    suggestUpstream=upstream, checkUpstream=tickUpstream)


class EditUpstreamBranch(RepoTask):
    def flow(self, localBranchName: str, remoteBranchName: str):
        # Bail if no-op
        currentUpstream = self.repo.branches.local[localBranchName].upstream
        currentUpstreamName = "" if not currentUpstream else currentUpstream.branch_name
        if remoteBranchName == currentUpstreamName:
            raise AbortTask()

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Upstreams
        self.repo.edit_upstream_branch(localBranchName, remoteBranchName)
        self.repoModel.prefs.setShadowUpstream(localBranchName, "")  # Clear any shadow upstream

        if remoteBranchName:
            self.epilog.status = _("Branch {0} now tracks {1}.", tquo(localBranchName), tquo(remoteBranchName))
        else:
            self.epilog.status = _("Branch {0} now tracks no upstream.", tquo(localBranchName))


class ResetHead(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn | TaskPrereqs.NoDetached

    def flow(self, onto: Oid):
        branchName = self.repo.head_branch_shorthand
        commitMessage = self.repo.get_commit_message(onto)
        submoduleDict = self.repo.listall_submodules_dict()
        hasSubmodules = bool(submoduleDict)

        dlg = ResetHeadDialog(onto, branchName, commitMessage, hasSubmodules, parent=self.parentWidget())
        dlg.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)  # don't leak dialog
        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        yield from self.flowDialog(dlg)

        resetMode = dlg.activeMode
        recurseSubmodules = dlg.recurseSubmodules()

        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Workdir

        modeArg = "--" + resetMode.name.lower()
        assert modeArg in ["--hard", "--mixed", "--soft"]

        yield from self.flowCallGit(
            "reset",
            modeArg,
            *argsIf(recurseSubmodules, "--recurse-submodules"),
            str(onto))

        self.epilog.status = _("Branch {0} was reset to {1} ({mode}).",
                               tquo(branchName), tquo(shortHash(onto)), mode=modeArg)


class FastForwardBranch(RepoTask):
    def flow(self, localBranchName: str = ""):
        if not localBranchName:
            self.checkPrereqs(TaskPrereqs.NoUnborn | TaskPrereqs.NoDetached)
            localBranchName = self.repo.head_branch_shorthand

        branch = self.repo.branches.local[localBranchName]
        upstream: Branch = branch.upstream
        if not upstream:
            raise AbortTask(_("Can’t fast-forward {0} because it isn’t tracking an upstream branch.",
                              bquo(branch.shorthand)))

        remoteBranchName = upstream.shorthand

        upToDate = yield from self._withGit(branch)

        ahead = False
        if upToDate:
            ahead = upstream.target != branch.target

        self.epilog.jumpTo = NavLocator.inRef(RefPrefix.HEADS + localBranchName)

        yield from self.flowEnterUiThread()
        if upToDate:
            message = paragraphs(
                _("No fast-forwarding necessary."),
                _("Your local branch {0} is ahead of {1}.") if ahead else
                _("Your local branch {0} is already up to date with {1}."))
            message = message.format(bquo(localBranchName), bquo(remoteBranchName))
            self.epilog.status = stripHtml(message)
            yield from self.flowConfirm(text=message, canCancel=False, dontShowAgainKey="NoFastForwardingNecessary")

    def onError(self, exc):
        if isinstance(exc, DivergentBranchesError):
            parentWidget = self.parentWidget()

            lb = exc.local_branch
            rb = exc.remote_branch
            text = paragraphs(
                _("Can’t fast-forward {0} to {1}.", bquo(lb.shorthand), bquo(rb.shorthand)),
                _("The branches are divergent."))
            qmb = showWarning(parentWidget, self.name(), text)

            # If it's the checked-out branch, suggest merging
            if lb.is_checked_out():
                mergeCaption = _("Merge into {0}", lquoe(lb.shorthand))
                mergeButton = qmb.addButton(mergeCaption, QMessageBox.ButtonRole.ActionRole)
                mergeButton.clicked.connect(lambda: MergeBranch.invoke(parentWidget, rb.name))
        else:
            super().onError(exc)

    def _withGit(self, branch: Branch):
        # Perform merge analysis with libgit2 first
        yield from self.flowEnterWorkerThread()
        upstream = branch.upstream
        analysis, _mergePref = self.repo.merge_analysis(branch.upstream.target, branch.name)

        if analysis & MergeAnalysis.UP_TO_DATE:
            # Local branch is up to date with remote branch, nothing to do.
            return True
        elif analysis == (MergeAnalysis.NORMAL | MergeAnalysis.FASTFORWARD):
            # Go ahead and fast-forward.
            pass
        elif analysis == MergeAnalysis.NORMAL:
            # Can't FF. Divergent branches?
            raise DivergentBranchesError(branch, upstream)
        else:
            # Unborn or something...
            raise NotImplementedError(f"Cannot fast-forward with {analysis!r}.")

        self.epilog.effects |= TaskEffects.Refs
        if branch.is_checked_out():
            self.epilog.effects |= TaskEffects.Head
            args = ["merge", "--ff-only", "--progress", branch.upstream_name]
        else:
            args = ["push", ".", f"{branch.upstream_name}:{branch.name}"]

        yield from self.flowEnterUiThread()
        driver = yield from self.flowCallGit(*args, autoFail=False)

        if driver.exitCode() != 0:
            raise DivergentBranchesError(branch, branch.upstream)

        return False


class MergeBranch(RepoTask):
    def flow(self, them: str | Oid, silentFastForward=False, autoFastForwardOptionName=""):
        # Figure out oid
        if isinstance(them, str):
            assert them.startswith("refs/")
            reference = self.repo.references[them]
            target = reference.target
            theirShorthand = reference.shorthand
            assert reference.type == ReferenceType.DIRECT
            assert isinstance(target, Oid)
        else:
            assert isinstance(them, Oid)
            target = them
            theirShorthand = shortHash(them)

        # Run merge analysis on background thread
        yield from self.flowEnterWorkerThread()
        self.repo.refresh_index()
        anyStagedFiles = self.repo.any_staged_changes
        anyConflicts = self.repo.any_conflicts
        myShorthand = self.repo.head_branch_shorthand
        analysis, pref = self.repo.merge_analysis(target)
        wantMergeCommit = True

        yield from self.flowEnterUiThread()
        logger.info(f"Merge analysis: {analysis!r} {pref!r}")

        if anyConflicts:
            message = paragraphs(
                _("Merging is not possible right now because you have unresolved conflicts."),
                _("Fix the conflicts to proceed."))
            raise AbortTask(message)

        elif anyStagedFiles:
            message = paragraphs(
                _("Merging is not possible right now because you have staged changes."),
                _("Commit your changes or stash them to proceed."))
            raise AbortTask(message)

        elif analysis == MergeAnalysis.UP_TO_DATE:
            message = paragraphs(
                _("No merge is necessary."),
                _("Your branch {0} is already up to date with {1}.", bquo(myShorthand), bquo(theirShorthand)))
            raise AbortTask(message, icon="information")

        elif analysis == MergeAnalysis.UNBORN:
            message = _("Cannot merge into an unborn head.")
            raise AbortTask(message)

        elif analysis == MergeAnalysis.FASTFORWARD | MergeAnalysis.NORMAL:
            if silentFastForward:
                wantMergeCommit = False
            else:
                wantMergeCommit = yield from self.confirmFastForward(myShorthand, theirShorthand, target, autoFastForwardOptionName)

        elif analysis == MergeAnalysis.NORMAL:
            title = _("Merging may cause conflicts")
            message = paragraphs(
                _("Merging {0} into {1} may cause conflicts.", bquo(theirShorthand), bquo(myShorthand)),
                _("You will need to fix the conflicts, if any. Then, commit the result to conclude the merge."))
            yield from self.flowConfirm(title=title, text=message, verb=_("Merge"),
                                        dontShowAgainKey="MergeMayCauseConflicts")

        else:
            raise NotImplementedError(f"Unsupported MergeAnalysis! ma={analysis!r} mp={pref!r}")

        # -----------------------------------------------------------
        # Actually perform the fast-forward or the merge

        yield from self._withGit(wantMergeCommit, theirShorthand)

        if wantMergeCommit:
            self.epilog.jumpTo = NavLocator.inWorkdir()
            self.epilog.status = _("Merging {0} into {1}.", tquo(theirShorthand), tquo(myShorthand))
        else:
            self.epilog.status = _("Branch {0} fast-forwarded to {1}.", tquo(myShorthand), tquo(theirShorthand))

    def confirmFastForward(self, myShorthand: str, theirShorthand: str, target: Oid, autoFastForwardOptionName: str):
        title = _("Fast-forwarding possible")
        message = _("Your branch {0} can simply be fast-forwarded to {1}.", bquo(myShorthand), bquo(theirShorthand))

        if autoFastForwardOptionName:
            notice = _("Automatic fast-forwarding is blocked by the {0} option in your Git config.", tquo("pull.ff"))
            message += f"<p><small>{notice}</small></p>"

        hint = paragraphs(
            _("<b>Fast-forwarding</b> means that the tip of your branch will be moved to a more "
              "recent commit in a linear path, without the need to create a merge commit."),
            _("In this case, {0} will be fast-forwarded to {1}.", bquo(myShorthand), bquo(shortHash(target))))

        actionButton = QPushButton(_("Create Merge Commit"))
        result = yield from self.flowConfirm(title=title, text=message, verb=_("Fast-Forward"),
                                             actionButton=actionButton, helpText=hint)

        # OK button is Fast-Forward; Merge is the other one. (Cancel would have aborted early.)
        # Also checking for Accepted so that unit tests can do qmb.accept().
        wantMergeCommit = result not in [QMessageBox.StandardButton.Ok, QDialog.DialogCode.Accepted]
        return wantMergeCommit

    def _withGit(self, wantMergeCommit: bool, theirShorthand: str):
        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Workdir

        driver = yield from self.flowCallGit(
            "merge",
            "--no-commit",
            "--no-edit",
            "--progress",
            "--verbose",
            *argsIf(wantMergeCommit, "--no-ff"),
            *argsIf(not wantMergeCommit, "--ff-only"),
            theirShorthand,
            autoFail=False  # don't abort the task if process returns non-0 (= conflicts)
        )

        if driver.exitCode() != 0:
            # If there are conflicts, "git merge" exits with nonzero and the
            # repo enters the MERGE state. If git exited with nonzero WITHOUT
            # entering the MERGE state, some other problem occurred.
            # (I guess we could infer this from the value of the exit code, but
            # it doesn't seem to be documented.)
            logger.warning(f"git merge rc={driver.exitCode()} - stderr: {driver.stderrScrollback()}")

            if self.repo.state() != RepositoryState.MERGE:
                raise AbortTask(driver.htmlErrorText())

        if wantMergeCommit:
            self.repoModel.prefs.draftCommitMessage = self.repo.message_without_conflict_comments


class RecallCommit(RepoTask):
    def flow(self):
        dlg = TextInputDialog(
            self.parentWidget(),
            _("Recall lost commit"),
            _("If you know the hash of a commit that isn’t part of any branches anymore, "
              "{app} will try to recall it for you.", app=qAppName()))
        dlg.okButton.setText(_("Recall"))

        yield from self.flowDialog(dlg)
        dlg.deleteLater()
        needle = dlg.lineEdit.text()

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Refs
        obj = self.repo[needle]
        commit: Commit = obj.peel(Commit)
        branchName = f"recall-{shortHash(commit.id)}"
        branchName = withUniqueSuffix(branchName, self.repo.listall_branches())
        self.repo.create_branch_from_commit(branchName, commit.id)
        self.epilog.jumpTo = NavLocator.inCommit(commit.id)



# Form implementation generated from reading ui file 'repostub.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_RepoStub(object):
    def setupUi(self, RepoStub):
        RepoStub.setObjectName("RepoStub")
        RepoStub.resize(670, 327)
        RepoStub.setWindowTitle("RepoStub")
        self.verticalLayout = QVBoxLayout(RepoStub)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.stackedWidget = QStackedWidget(parent=RepoStub)
        self.stackedWidget.setObjectName("stackedWidget")
        self.promptPage = QWidget()
        self.promptPage.setObjectName("promptPage")
        self.gridLayout = QGridLayout(self.promptPage)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QSpacerItem(20, 132, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem, 0, 1, 1, 1)
        self.promptIcon = QLabel(parent=self.promptPage)
        self.promptIcon.setText("ICON")
        self.promptIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.promptIcon.setObjectName("promptIcon")
        self.gridLayout.addWidget(self.promptIcon, 1, 1, 1, 1)
        self.promptNameLabel = QLabel(parent=self.promptPage)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.promptNameLabel.sizePolicy().hasHeightForWidth())
        self.promptNameLabel.setSizePolicy(sizePolicy)
        self.promptNameLabel.setMaximumSize(QSize(300, 16777215))
        self.promptNameLabel.setText("Long Long Long Long Long Long Long Long Long Long Name")
        self.promptNameLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.promptNameLabel.setWordWrap(True)
        self.promptNameLabel.setObjectName("promptNameLabel")
        self.gridLayout.addWidget(self.promptNameLabel, 2, 1, 1, 1)
        self.promptReadyLabel = QLabel(parent=self.promptPage)
        self.promptReadyLabel.setMinimumSize(QSize(300, 0))
        self.promptReadyLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.promptReadyLabel.setWordWrap(True)
        self.promptReadyLabel.setObjectName("promptReadyLabel")
        self.gridLayout.addWidget(self.promptReadyLabel, 3, 1, 1, 1)
        spacerItem1 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem1, 4, 0, 1, 1)
        self.promptLoadButton = QPushButton(parent=self.promptPage)
        self.promptLoadButton.setObjectName("promptLoadButton")
        self.gridLayout.addWidget(self.promptLoadButton, 4, 1, 1, 1)
        spacerItem2 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem2, 4, 2, 1, 1)
        spacerItem3 = QSpacerItem(20, 132, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem3, 5, 1, 1, 1)
        self.stackedWidget.addWidget(self.promptPage)
        self.progressPage = QWidget()
        self.progressPage.setObjectName("progressPage")
        self.gridLayout_2 = QGridLayout(self.progressPage)
        self.gridLayout_2.setObjectName("gridLayout_2")
        spacerItem4 = QSpacerItem(20, 154, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout_2.addItem(spacerItem4, 0, 1, 1, 1)
        self.progressLabel = QLabel(parent=self.progressPage)
        self.progressLabel.setMinimumSize(QSize(300, 0))
        self.progressLabel.setText("PROGRESS MESSAGE Long Long Long Long Long Long Long Long Long Long Name")
        self.progressLabel.setWordWrap(True)
        self.progressLabel.setObjectName("progressLabel")
        self.gridLayout_2.addWidget(self.progressLabel, 1, 1, 1, 2)
        spacerItem5 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem5, 2, 0, 1, 1)
        self.progressBar = QProgressBar(parent=self.progressPage)
        self.progressBar.setTextVisible(False)
        self.progressBar.setObjectName("progressBar")
        self.gridLayout_2.addWidget(self.progressBar, 2, 1, 1, 1)
        self.progressAbortButton = QPushButton(parent=self.progressPage)
        self.progressAbortButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.progressAbortButton.setText("")
        self.progressAbortButton.setObjectName("progressAbortButton")
        self.gridLayout_2.addWidget(self.progressAbortButton, 2, 2, 1, 1)
        spacerItem6 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem6, 2, 3, 1, 1)
        spacerItem7 = QSpacerItem(20, 154, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout_2.addItem(spacerItem7, 3, 1, 1, 1)
        self.stackedWidget.addWidget(self.progressPage)
        self.verticalLayout.addWidget(self.stackedWidget)

        self.retranslateUi(RepoStub)
        self.stackedWidget.setCurrentIndex(0)

    def retranslateUi(self, RepoStub):
        self.promptReadyLabel.setText(_p("RepoStub", "Ready to load this repository."))
        self.promptLoadButton.setText(_p("RepoStub", "Load"))
        self.progressAbortButton.setToolTip(_p("RepoStub", "Abort"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import re
from collections.abc import Callable
from contextlib import suppress
from pathlib import Path
from typing import ClassVar

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.forms.commitinfodialog import CommitInfoDialog
from gitfourchette.forms.ignorepatterndialog import IgnorePatternDialog
from gitfourchette.forms.reposettingsdialog import RepoSettingsDialog
from gitfourchette.gitdriver.parsers import iterateLines
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import Oid, Signature
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID, BEGIN_SSH_SIGNATURE, GpgStatus
from gitfourchette.tasks import TaskEffects
from gitfourchette.tasks.repotask import RepoTask, AbortTask
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class EditRepoSettings(RepoTask):
    def flow(self):
        repo = self.repo
        repoPrefs = self.repoModel.prefs

        dlg = RepoSettingsDialog(repo, repoPrefs, self.parentWidget())
        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        yield from self.flowDialog(dlg)

        localName, localEmail = dlg.localIdentity()
        nickname = dlg.ui.nicknameEdit.text()
        customKeyFile = dlg.ui.keyFilePicker.privateKeyPath()
        dlg.deleteLater()

        configObject = repo.config
        for key, value in [("user.name", localName), ("user.email", localEmail)]:
            if value:
                configObject[key] = value
            else:
                with suppress(KeyError):
                    del configObject[key]
        repo.scrub_empty_config_section("user")

        if customKeyFile != repoPrefs.customKeyFile:
            repoPrefs.customKeyFile = customKeyFile
            repoPrefs.setDirty()

        if nickname != settings.history.getRepoNickname(self.repo.workdir, strict=True):
            settings.history.setRepoNickname(self.repo.workdir, nickname)
            settings.history.setDirty()
            self.rw.nameChange.emit()

        repoPrefs.write()


class GetCommitInfo(RepoTask):
    @staticmethod
    def formatSignature(sig: Signature):
        dateText = signatureDateFormat(sig)
        return f"{escape(sig.name)} &lt;{escape(sig.email)}&gt;<br><small>{escape(dateText)}</small>"

    def defaultJumpCallback(self, locator: NavLocator):
        self.epilog.jumpTo = locator

    def flow(
            self,
            oid: Oid,
            withDebugInfo: bool = False,
            jumpCallback: Callable[[NavLocator], None] | None = None,
    ):
        if jumpCallback is None:
            jumpCallback = self.defaultJumpCallback

        linkBundle = DocumentLinks()

        def commitLink(commitId):
            if commitId == UC_FAKEID:
                locator = NavLocator.inWorkdir()
            else:
                locator = NavLocator.inCommit(commitId)
            link = linkBundle.new(lambda: jumpCallback(locator))
            html = linkify(shortHash(commitId), link)
            return html

        def tableRow(th, td):
            colon = _(":")
            return f"<tr><th>{th}{colon}</th><td>{td}</td></tr>"

        repo = self.repo
        repoModel = self.repoModel
        commit = repo.peel_commit(oid)

        # Break down commit message into summary/details
        summary, contd = messageSummary(commit.message)
        details = commit.message if contd else ""

        # Parent commits
        parentHashes = [commitLink(p) for p in commit.parent_ids]
        numParents = len(parentHashes)
        parentTitle = _n("Parent", "{n} Parents", numParents)
        if numParents > 0:
            parentMarkup = ', '.join(parentHashes)
        elif not repo.is_shallow:
            parentMarkup = "-"
        else:
            shallowCloneBlurb = _("You’re working in a shallow clone. This commit may actually have parents in the full history.")
            parentMarkup = tagify(shallowCloneBlurb, "<p><em>")

        # Committer
        if commit.author == commit.committer:
            committerMarkup = tagify(_("(same as author)"), "<i>")
        else:
            committerMarkup = self.formatSignature(commit.committer)

        # GPG
        gpgStatus, gpgKeyInfo = self.repoModel.getCachedGpgStatus(commit)
        if gpgStatus == GpgStatus.Unsigned:
            gpgMarkup = tagify(_("(not signed)"), "<i>")
        elif gpgKeyInfo:
            gpgMarkup = f"{gpgStatus.iconHtml()} {trtables.enum(gpgStatus)}<br><small>{escape(gpgKeyInfo)}</small>"
        else:
            gpgMarkup = f"{gpgStatus.iconHtml()} {trtables.enum(gpgStatus)}"

        # Assemble table rows
        table = tableRow(_("Hash"), commit.id)
        table += tableRow(parentTitle, parentMarkup)
        table += tableRow(_("Author"), self.formatSignature(commit.author))
        table += tableRow(_("Committer"), committerMarkup)
        table += tableRow(_("Signature"), gpgMarkup)

        # Graph debug info
        if withDebugInfo:
            graph = repoModel.graph
            seqIndex = graph.getCommitRow(oid)
            frame = graph.getFrame(seqIndex)
            homeChain = frame.homeChain()
            homeChainTopId = graph.getFrame(int(homeChain.topRow)).commit
            homeChainTopStr = commitLink(homeChainTopId) if type(homeChainTopId) is Oid else str(homeChainTopId)
            table += tableRow("Graph row", repr(graph.commitRows[oid]))
            table += tableRow("Home chain", f"{homeChain.topRow!r} {homeChainTopStr} ({id(homeChain) & 0xFFFFFFFF:X})")
            table += tableRow("Arcs", f"{len(frame.openArcs)} open, {len(frame.solvedArcs)} solved")
            # table += tableRow("View row", self.rw.graphView.currentIndex().row())
            details = str(frame) + "\n\n" + details

        title = _("Commit info: {0}", shortHash(commit.id))

        markup = f"""\
        <style>
            table {{ margin-top: 16px; }}
            th, td {{ padding-bottom: 4px; }}
            th {{
                text-align: right;
                padding-right: 8px;
                font-weight: normal;
                white-space: pre;
                color: {mutedTextColorHex(self.parentWidget())};
            }}
        </style>
        <big>{summary}</big>
        <table>{table}</table>
        """

        dialog = CommitInfoDialog(self.parentWidget(), title, markup, details)

        dialog.summaryLabel.linkActivated.connect(linkBundle.processLink)
        dialog.summaryLabel.linkActivated.connect(dialog.accept)

        yield from self.flowDialog(dialog)


class VerifyGpgSignature(RepoTask):
    _GnupgStatusTable: ClassVar = {
        # The order in this table is significant for parseGnupgVerification
        "NO_PUBKEY" : GpgStatus.MissingKey,
        "GOODSIG"   : GpgStatus.GoodUntrusted,
        "EXPSIG"    : GpgStatus.ExpiredSig,
        "EXPKEYSIG" : GpgStatus.ExpiredKey,
        "REVKEYSIG" : GpgStatus.RevokedKey,
        "KEYREVOKED": GpgStatus.RevokedKey,
        "BADSIG"    : GpgStatus.Bad,
    }

    _GnupgLinePattern: ClassVar = re.compile(
        r"^\[GNUPG:]\s+(\w+)\s*(.*)$")

    _SshGoodTrustedPattern: ClassVar = re.compile(
        r'^Good "git" signature for (.+) with \S+ key (.+)')

    _SshGoodUntrustedPattern: ClassVar = re.compile(
        r'^Good "git" signature with \S+ key (.+)')

    def flow(self, oid: Oid, dialogParent: QWidget | None = None):
        commit = self.repo.peel_commit(oid)

        gpgSignature, _gpgPayload = commit.gpg_signature
        if not gpgSignature:
            raise AbortTask(_("Commit {0} is not signed, so it cannot be verified.", tquo(shortHash(oid))))

        isSsh = gpgSignature.startswith(BEGIN_SSH_SIGNATURE)

        driver = yield from self.flowCallGit("verify-commit", "--raw", str(oid), autoFail=False)
        fail = driver.exitCode() != 0
        stderr = driver.stderrScrollback()

        status, keyInfo = self.parseScrollback(stderr, isSsh)

        # Update gpg status cache
        self.repoModel.cacheGpgStatus(oid, status, keyInfo)

        paras = [f"{stockIconImgTag(status.iconName())} {trtables.enum(status)}"]

        if status == GpgStatus.MissingKey:
            paras.append(_("Hint: Public key {0} isn’t in your GPG keyring. "
                           "You can try to import it from a trusted source, "
                           "then verify this commit again.", f"<em>{escape(keyInfo)}</em>"))

        if keyInfo:
            paras.append(f"<i>{escape(keyInfo)}</i>")

        title = _("Verify signature in commit {0}", tquo(shortHash(oid)))

        mbIcon: MessageBoxIconName
        if not fail:
            mbIcon = "information"
        elif status in [GpgStatus.RevokedKey, GpgStatus.Bad]:
            mbIcon = "critical"
        else:
            mbIcon = "warning"

        qmb = asyncMessageBox(self.parentWidget(), mbIcon, title, paragraphs(*paras),
                              QMessageBox.StandardButton.Ok)# | QMessageBox.StandardButton.Help)
        qmb.setDetailedText(driver.stderrScrollback())

        if keyInfo:
            hintButton = QPushButton(qmb)
            qmb.addButton(hintButton, QMessageBox.ButtonRole.HelpRole)
            hintButton.setText(_("Copy &Key ID"))
            hintButton.clicked.disconnect()  # Qt internally wires the button to close the dialog; undo that.

            def onCopyClicked():
                QApplication.clipboard().setText(keyInfo)
                QToolTip.showText(QCursor.pos(), _("{0} copied to clipboard.", tquo(keyInfo)))
            hintButton.clicked.connect(onCopyClicked)

            # Adding an extra button seems to remove the OK button's Escape key mapping
            qmb.setEscapeButton(qmb.button(QMessageBox.StandardButton.Ok))

        yield from self.flowDialog(qmb)

    @classmethod
    def parseScrollback(cls, scrollback: str, isSsh: bool) -> tuple[GpgStatus, str]:
        if isSsh:
            return cls.parseSshVerification(scrollback)
        else:
            return cls.parseGnupgVerification(scrollback)

    @classmethod
    def parseGnupgVerification(cls, scrollback: str) -> tuple[GpgStatus, str]:
        # https://github.com/gpg/gnupg/blob/master/doc/DETAILS#general-status-codes
        # https://github.com/git/git/blob/v2.51.0/gpg-interface.c#L184

        report = {}
        for line in scrollback.splitlines():
            match = cls._GnupgLinePattern.match(line)
            if not match:
                continue
            token = match.group(1)
            blurb = match.group(2)
            report[token] = blurb

        # Find the most significant status.
        # The order of the keys in _GnupgStatusTable is significant!
        status = next((status for token, status in cls._GnupgStatusTable.items() if token in report),
                      GpgStatus.CantCheck)

        # Bump GoodUntrusted to GoodTrusted if we trust this
        if status == GpgStatus.GoodUntrusted and ("TRUST_ULTIMATE" in report) or ("TRUST_FULLY" in report):
            status = GpgStatus.GoodTrusted

        # Try to extract a key ID or fingerprint for informative purposes
        keyInfo = ""
        for token in cls._GnupgStatusTable:
            keyInfo = report.get(token, "")
            if keyInfo:
                break

        return status, keyInfo

    @classmethod
    def parseSshVerification(cls, scrollback: str) -> tuple[GpgStatus, str]:
        lines = scrollback.splitlines()

        if not lines:
            return GpgStatus.CantCheck, ""

        match = cls._SshGoodTrustedPattern.match(lines[0])
        if match:
            assert "No principal matched." not in lines
            keyInfo = f"{match.group(1)} {match.group(2)}"
            return GpgStatus.GoodTrusted, keyInfo

        match = cls._SshGoodUntrustedPattern.match(lines[0])
        if match:
            assert "No principal matched." in lines
            keyInfo = _("(Not found in SSH allowed signers file)") + " " + match.group(1)
            return GpgStatus.GoodUntrusted, keyInfo

        if lines[0].startswith("Could not verify signature."):
            return GpgStatus.Bad, ""

        return GpgStatus.CantCheck, ""


class VerifyGpgQueue(RepoTask):
    def isFreelyInterruptible(self) -> bool:
        return True

    def flow(self):
        self.epilog.effects = TaskEffects.Nothing
        graphView = self.rw.graphView
        repoModel = self.repoModel

        while repoModel.gpgVerifyQueue:
            oid = repoModel.gpgVerifyQueue.pop()

            currentStatus, keyInfo = repoModel.gpgStatusCache.get(oid, (GpgStatus.Unsigned, ""))
            if currentStatus != GpgStatus.Pending:
                continue

            try:
                visibleIndex = self.visibleIndex(oid)
            except LookupError:
                continue

            try:
                driver = yield from self.flowCallGit("verify-commit", "--raw", str(oid), autoFail=False)
            except AbortTask:
                # This task may be issued repeatedly.
                # Don't let AbortTask spam dialog boxes if git failed to start.
                repoModel.cacheGpgStatus(oid, GpgStatus.ProcessError)
                graphView.update(visibleIndex)
                continue

            stderr = driver.stderrScrollback()
            isSsh = keyInfo == "ssh"

            status, keyInfo = VerifyGpgSignature.parseScrollback(stderr, isSsh)
            repoModel.cacheGpgStatus(oid, status, keyInfo)
            graphView.update(visibleIndex)

    def visibleIndex(self, oid: Oid) -> QModelIndex:
        graphView = self.rw.graphView

        index = graphView.getFilterIndexForCommit(oid)  # raises SelectCommitError (LookupError) if hidden
        assert index.isValid()

        top = graphView.indexAt(QPoint_zero)
        bottom = graphView.indexAt(graphView.rect().bottomLeft())

        topRow = top.row() if top.isValid() else 0
        bottomRow = bottom.row() if bottom.isValid() else 0x3FFFFFFF

        if topRow <= index.row() <= bottomRow:
            return index

        raise IndexError()


class NewIgnorePattern(RepoTask):
    def flow(self, seedPath: str):
        dlg = IgnorePatternDialog(seedPath, self.parentWidget())
        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        yield from self.flowDialog(dlg)

        pattern = dlg.pattern
        if not pattern:
            raise AbortTask()

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Workdir

        relativeExcludePath = dlg.excludePath
        excludePath = Path(self.repo.in_workdir(relativeExcludePath))

        # Read existing exclude text
        excludeText = ""
        if excludePath.exists():
            excludeText = excludePath.read_text("utf-8")
            if not excludeText.endswith("\n"):
                excludeText += "\n"

        excludeText += pattern + "\n"
        excludePath.write_text(excludeText, "utf-8")

        self.epilog.status = _("Added to {file}: {pattern}", pattern=pattern, file=relativeExcludePath)

        # Jump to .gitignore
        if self.repo.is_in_workdir(str(excludePath)):
            self.epilog.jumpTo = NavLocator.inUnstaged(str(excludePath))


class QueryCommitsTouchingPath(RepoTask):
    """
    Resolve commits reachable from any refs that modify the given pathspec
    (via git log).
    """

    def broadcastProcesses(self) -> bool:
        return False

    def canKill(self, task: RepoTask) -> bool:
        # Can kill same class to replace stale query
        return isinstance(task, QueryCommitsTouchingPath)

    def flow(self, pathspec: str):
        cpf = self.repoModel.commitPathspecFilter
        cpf.clear()
        cpf.needle = pathspec

        oids = yield from self._findMatchingCommits(pathspec)

        # Needle shouldn't have been tampered with while we were waiting on git
        assert cpf.needle == pathspec

        # Evict commits we don't have in memory
        oids.intersection_update(self.repoModel.graph.commitRows)

        self.epilog.status = _n(
            "Found {n} commit touching {path}{where}.",
            "Found {n} commits touching {path}{where}.",
            n=len(oids), path=tquo(pathspec),
            where=_(" in the truncated history") if self.repoModel.truncatedHistory else "")

        if (self.repoModel.hiddenCommits
                and any(o in self.repoModel.hiddenCommits for o in oids)):
            self.epilog.status += _(" More results in hidden branches.")

        cpf.matchingIds = oids

        # Cache workdir status
        if self.repoModel.workdirMatchesPathNeedle(pathspec):
            cpf.matchingIds.add(UC_FAKEID)

        cpf.resultsUpdated.emit()

    def _findMatchingCommits(self, pathspec: str) -> RepoTask.Flow[set[Oid]]:
        oids: set[Oid] = set()

        pathspec = pathspec.strip()
        if not pathspec:
            return oids

        driver = yield from self.flowCallGit(
            "-c", "core.abbrev=no",
            "--icase-pathspecs",  # same as ':(icase)' prefix in pathspec
            "log", "--all", "--format=%H",
            "--show-pulls",  # keep relevant merge commits
            "--", pathspec,
            autoFail=False,
        )

        stdout = driver.stdoutScrollback().rstrip()
        for pos, endPos in iterateLines(stdout):
            line = stdout[pos: endPos].rstrip()
            if not line:
                continue
            assert len(line) == 40
            oid = Oid(hex=line)
            oids.add(oid)
        return oids



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
import functools
import logging
from collections.abc import Sequence, Iterable, Callable, Set
from typing import cast

from gitfourchette.graph.graph import Graph, BatchRow, KF_INTERVAL, Oid, CommitTraits
from gitfourchette.graph.graphsplicer import GraphSplicer
from gitfourchette.graph.graphtrickle import GraphTrickle
from gitfourchette.graph.graphweaver import GraphWeaver
from gitfourchette.toolbox import Benchmark

logger = logging.getLogger(__name__)


def _ensureSet(x) -> set:
    if x is None:
        return set()
    assert type(x) is not dict
    if type(x) is not set:
        return set(x)
    assert type(x) is set
    return x


@dataclasses.dataclass
class MockCommit:
    id: Oid
    parent_ids: Sequence[Oid]


# pygit2.Oid can't be subclassed, and it prevents reassigning __repr__,
# so we have to create a separate class that honors the Oid protocol.
@functools.total_ordering  # saves me from implementing le,ne,gt,ge
class MockOid:
    """
    An Oid-like object augmented with a custom __repr__ string for unit tests.

    MockOid("hello") encodes "hello" into a 20-byte buffer that serves as a
    SHA-1 hash for interoperability with code that expects a real Oid.
    MockOid.__repr__ decodes the original string from the 20-byte buffer
    (e.g. "hello").
    """
    @classmethod
    def encodeAll(cls, texts: Iterable[str]) -> list[Oid]:
        assert not isinstance(texts, str), "don't pass a single string"
        return [cast(Oid, MockOid(o)) for o in texts]

    def __init__(self, text: str):
        encoded: bytes = text.encode()
        assert len(encoded) <= 20, f"too long for 20-byte Oid: {text}"
        self.oid = Oid(raw=encoded)

    @property
    def __class__(self):
        return Oid

    @__class__.setter
    def __class__(self, value):
        raise TypeError("cannot reassign __class__")

    @property
    def raw(self) -> bytes:
        return self.oid.raw

    def __str__(self):
        return self.oid.raw.rstrip(b'\0').decode()

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.oid)

    def __lt__(self, other):
        if isinstance(other, str):
            return str(self) < other
        if isinstance(other, MockOid):
            other = other.oid
        return self.oid.__lt__(other)

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        if isinstance(other, MockOid):
            other = other.oid
        return self.oid.__eq__(other)


class GraphBuildLoop:
    onKeyframe: Callable[[int], None]

    def __init__(
            self,
            heads=None,
            hideSeeds=None,
            localSeeds=None,
            forceHide=None,
            keyframeInterval=KF_INTERVAL
    ):
        heads = _ensureSet(heads)
        hideSeeds = _ensureSet(hideSeeds)
        # If localSeeds was omitted, all heads are local by default
        localSeeds = _ensureSet(heads if localSeeds is None else localSeeds)

        self.graph, self.weaver = GraphWeaver.newGraph()
        self.hiddenTrickle = GraphTrickle.newHiddenTrickle(heads, hideSeeds, forceHide)
        self.foreignTrickle = GraphTrickle.newForeignTrickle(heads, localSeeds)
        self.keyframeInterval = keyframeInterval

        self.onKeyframe = GraphBuildLoop.defaultOnKeyframe

    def sendAll(self, sequence):
        gen = self.coBuild()
        gen.send(None)  # prime it
        for c in sequence:
            gen.send(c)
        gen.close()
        return self

    @staticmethod
    def defaultOnKeyframe(i: int):
        pass

    def coBuild(self):
        graph = self.graph
        weaver = self.weaver
        hiddenTrickle = self.hiddenTrickle
        foreignTrickle = self.foreignTrickle
        keyframeInterval = self.keyframeInterval

        while True:
            try:
                commit = yield
            except GeneratorExit:
                break

            oid = commit.id
            parents = commit.parent_ids

            weaver.newCommit(oid, parents)
            hiddenTrickle.newCommit(oid, parents)
            foreignTrickle.newCommit(oid, parents)

            row = weaver.row
            rowInt = int(row)
            assert type(row) is BatchRow
            assert rowInt >= 0
            graph.commitRows[oid] = row

            # Save keyframes at regular intervals for faster random access.
            if rowInt % keyframeInterval == 0:
                graph.saveKeyframe(weaver)
                self.onKeyframe(rowInt)

        logger.debug(f"Peak arc count: {weaver.peakArcCount}")

    @property
    def hiddenCommits(self):
        return self.hiddenTrickle.flaggedSet

    @property
    def foreignCommits(self):
        return self.foreignTrickle.flaggedSet


class GraphSpliceLoop:
    def __init__(
            self,
            graph: Graph,
            oldCommitSequence: list[CommitTraits],
            oldHeads: Iterable[Oid],
            newHeads: Iterable[Oid],
            hideSeeds: Set[Oid] | None = None,
            localSeeds: Set[Oid] | None = None,
            keyframeInterval=KF_INTERVAL,
    ):
        oldHeads = _ensureSet(oldHeads)
        newHeads = _ensureSet(newHeads)
        hideSeeds = _ensureSet(hideSeeds)
        # If localSeeds was omitted, all heads are local by default
        localSeeds = _ensureSet(newHeads if localSeeds is None else localSeeds)

        self.graph = graph
        self.oldCommitSequence = oldCommitSequence
        self.commitSequence: list[CommitTraits] = []  # unknown yet
        self.oldHeads = oldHeads
        self.newHeads = newHeads
        self.hideSeeds = hideSeeds
        self.localSeeds = localSeeds
        self.keyframeInterval = keyframeInterval

        self.splicer = GraphSplicer(self.graph, self.oldHeads, self.newHeads)
        self.hiddenTrickle = GraphTrickle.newHiddenTrickle(self.newHeads, self.hideSeeds)
        self.foreignTrickle = GraphTrickle.newForeignTrickle(self.newHeads, self.localSeeds)

        self.numRowsRemoved = 0
        self.numRowsAdded = 0

    def sendAll(self, sequence):
        gen = self.coSplice()
        gen.send(None)  # prime it
        for c in sequence:
            try:
                gen.send(c)
            except StopIteration:
                break
        gen.close()
        return self

    def coSplice(self):
        newCommitSequence = []
        splicer = self.splicer
        hiddenTrickle = self.hiddenTrickle
        foreignTrickle = self.foreignTrickle

        while not splicer.done:
            try:
                commit = yield
            except GeneratorExit:
                splicer.onGraphDepleted()
                break
            oid = commit.id
            parents = commit.parent_ids

            newCommitSequence.append(commit)
            splicer.spliceNewCommit(oid, parents, self.keyframeInterval)
            hiddenTrickle.newCommit(oid, parents)
            foreignTrickle.newCommit(oid, parents)

        if splicer.foundEquilibrium:
            nRemoved = splicer.equilibriumOldRow
            nAdded = splicer.equilibriumNewRow
        else:
            nRemoved = -1  # We could use len(self.commitSequence), but -1 will force refreshRepo to replace the model wholesale
            nAdded = len(newCommitSequence)

        # Piece correct commit sequence back together
        with Benchmark("Reassemble commit sequence"):
            if not splicer.foundEquilibrium:
                pass  # keep newCommitSequence
            elif nAdded == 0 and nRemoved == 0:
                newCommitSequence = self.oldCommitSequence
            elif nRemoved == 0:
                newCommitSequence = newCommitSequence[:nAdded] + self.oldCommitSequence
            else:
                newCommitSequence = newCommitSequence[:nAdded] + self.oldCommitSequence[nRemoved:]

        # Finish patching hidden/foreign commit sets.
        # Keep feeding commits to trickle until it stabilizes.
        if splicer.foundEquilibrium:
            with Benchmark("Stabilize trickles"):
                row = nAdded + 1
                r1 = self._stabilizeTrickle(hiddenTrickle, row, newCommitSequence)
                r2 = self._stabilizeTrickle(foreignTrickle, row, newCommitSequence)
            logger.debug(f"Trickle stabilization: Hidden={r1}; Foreign={r2}")

        self.numRowsRemoved = nRemoved
        self.numRowsAdded = nAdded
        self.commitSequence = newCommitSequence

    @staticmethod
    def _stabilizeTrickle(trickle: GraphTrickle, startRow: int, newCommitSequence: list[MockCommit]):
        if trickle.done:
            return startRow

        for row in range(startRow, len(newCommitSequence)):
            commit = newCommitSequence[row]
            trickle.newCommit(commit.id, commit.parent_ids)

            # See if finished (not too often - expensive)
            if ((row & 0xFF) == 0) and trickle.done:
                return row  # type: ignore[unreachable]

    @property
    def hiddenCommits(self):
        return self.hiddenTrickle.flaggedSet

    @property
    def foreignCommits(self):
        return self.foreignTrickle.flaggedSet



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import os

from gitfourchette.qt import *


class PersistentFileDialog:
    @staticmethod
    def getPath(key: str, fallbackPath: str = ""):
        from gitfourchette import settings
        return settings.history.fileDialogPaths.get(key, fallbackPath)

    @staticmethod
    def savePath(key, path):
        if path:
            from gitfourchette import settings
            settings.history.fileDialogPaths[key] = path
            settings.history.write()

    @staticmethod
    def install(qfd: QFileDialog, key: str):
        # Don't use native dialog in unit tests
        # (macOS native file dialog cannot be controlled from unit tests)
        qfd.setOption(QFileDialog.Option.DontUseNativeDialog, APP_TESTMODE)

        # Restore saved path
        savedPath = PersistentFileDialog.getPath(key)
        if savedPath:
            savedPath = os.path.dirname(savedPath)
            if os.path.exists(savedPath):
                qfd.setDirectory(savedPath)

        # Remember selected path
        qfd.fileSelected.connect(lambda path: PersistentFileDialog.savePath(key, path))

        return qfd

    @staticmethod
    def saveFile(parent: QWidget, key: str, caption: str, initialFilename="", filter="", selectedFilter=""):
        qfd = QFileDialog(parent, caption, initialFilename, filter)
        qfd.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        qfd.setFileMode(QFileDialog.FileMode.AnyFile)
        if selectedFilter:
            qfd.selectNameFilter(selectedFilter)
        qfd.setWindowModality(Qt.WindowModality.WindowModal)
        PersistentFileDialog.install(qfd, key)
        return qfd

    @staticmethod
    def openFile(parent: QWidget, key: str, caption: str, fallbackPath="", filter="", selectedFilter=""):
        qfd = QFileDialog(parent, caption, fallbackPath, filter)
        qfd.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        qfd.setFileMode(QFileDialog.FileMode.AnyFile)
        if selectedFilter:
            qfd.selectNameFilter(selectedFilter)
        qfd.setWindowModality(Qt.WindowModality.WindowModal)
        PersistentFileDialog.install(qfd, key)
        return qfd

    @staticmethod
    def openDirectory(parent: QWidget, key: str, caption: str, options=QFileDialog.Option.ShowDirsOnly):
        qfd = QFileDialog(parent, caption)
        qfd.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        qfd.setFileMode(QFileDialog.FileMode.Directory)
        qfd.setOptions(options)
        qfd.setWindowModality(Qt.WindowModality.WindowModal)
        PersistentFileDialog.install(qfd, key)
        return qfd



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import gc
import logging

import pygit2

from gitfourchette.qt import *
from gitfourchette.toolbox.qtutils import setFontFeature

logger = logging.getLogger(__name__)


class MemoryIndicator(QPushButton):
    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.setObjectName("MemoryIndicator")
        self.setText("Memory")

        # No border: don't let it thicken the status bar
        self.setStyleSheet("border: none; text-align: right; padding-right: 8px;")

        font: QFont = self.font()
        font.setPointSize(font.pointSize() * 85 // 100)
        setFontFeature(font, "tnum")  # Tabular numbers
        self.setFont(font)

        width = 220

        self.setMinimumWidth(width)
        self.setMaximumWidth(width)
        self.clicked.connect(self.onMemoryIndicatorClicked)
        self.setToolTip("Force GC")
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.lastUpdate = 0

    def onMemoryIndicatorClicked(self):
        gc.collect()

        windows = '\n'.join(f'\t* {w.__class__.__name__} {w.objectName()}' for w in QApplication.topLevelWindows())
        widgets = '\n'.join(f'\t* {w.__class__.__name__} {w.objectName()}' for w in QApplication.topLevelWidgets())
        report = f"\nTop-Level Windows:\n{windows}\nTop-Level Widgets:\n{widgets}\n"
        logger.info(report)

        self.lastUpdate = 0
        self.updateMemoryIndicator()

    def paintEvent(self, event: QPaintEvent):
        self.updateMemoryIndicator()
        super().paintEvent(event)

    def updateMemoryIndicator(self):
        now = QDateTime.currentMSecsSinceEpoch()
        if now - self.lastUpdate < 30:
            return
        self.lastUpdate = now

        allIds = set()
        for w in QApplication.topLevelWidgets():
            allIds.add(id(w))
            allIds.update(id(o) for o in w.findChildren(QObject))
        numQObjects = len(allIds)

        cacheMem, _dummy = pygit2.settings.cached_memory
        fds = QLocale().formattedDataSize(cacheMem, 0, QLocale.DataSizeFormat.DataSizeSIFormat)
        self.setText(f"git: {fds}    qto: {numQObjects}    pyo: {len(gc.get_objects())}")



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

import logging
import os
from dataclasses import dataclass
from typing import Literal, ClassVar
from weakref import ReferenceType

from gitfourchette import colors
from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.codeview.codewindow import CodeWindow
from gitfourchette.exttools.mergedriver import MergeDriver
from gitfourchette.exttools.toolprocess import ToolProcess
from gitfourchette.forms.ui_conflictview import Ui_ConflictView
from gitfourchette.gitdriver import GitConflict, GitConflictSides
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.tasks import HardSolveConflicts, AcceptMergeConflictResolution, OpenMergeTool
from gitfourchette.tasks.indextasks import PreviewDeltaFile
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class ConflictView(QWidget):
    openPrefs = Signal(str)

    repoModel: RepoModel
    currentConflict: GitConflict | None
    currentMerge: MergeDriver | None
    currentMergeState: MergeDriver.State
    currentPreviewWindows: list[ReferenceType[QWidget]]

    def __init__(self, repoModel: RepoModel, parent=None):
        super().__init__(parent)

        self.repoModel = repoModel
        self.currentConflict = None
        self.currentMerge = None
        self.currentMergeState = MergeDriver.State.Idle
        self.currentPreviewWindows = []

        self.ui = Ui_ConflictView()
        self.ui.setupUi(self)

        self.setBackgroundRole(QPalette.ColorRole.Base)
        self.setAutoFillBackground(True)

        tweakWidgetFont(self.ui.titleLabel, 130)
        tweakWidgetFont(self.ui.mergeToolButton, 88)

        self.ui.oursPreviewButton.setIcon(stockIcon("view-visible"))
        self.ui.theirsPreviewButton.setIcon(stockIcon("view-visible"))
        self.ui.oursPreviewButton.clicked.connect(lambda: self.openPreview("ours"))
        self.ui.theirsPreviewButton.clicked.connect(lambda: self.openPreview("theirs"))

        self.ui.mergeToolButton.clicked.connect(lambda: self.openPrefs.emit(ToolProcess.PrefKeyMergeTool))
        self.ui.oursButton.clicked.connect(lambda: self.execute("ours"))
        self.ui.theirsButton.clicked.connect(lambda: self.execute("theirs"))
        self.ui.confirmDeletionButton.clicked.connect(lambda: self.execute("ancestor"))
        self.ui.mergeButton.clicked.connect(lambda: self.execute("merge"))
        self.ui.confirmMergeButton.clicked.connect(self.confirmMergeResolution)
        self.ui.discardMergeButton.clicked.connect(self.discardMergeResolution)
        self.ui.reworkMergeButton.clicked.connect(lambda: self.execute("remerge"))
        self.ui.cancelMergeInProgress.clicked.connect(self.cancelMergeInProgress)

        self.ui.confirmMergeButton.setIcon(stockIcon("SP_DialogSaveButton"))
        self.ui.discardMergeButton.setIcon(stockIcon("SP_DialogDiscardButton"))
        self.ui.reworkMergeButton.setIcon(stockIcon("SP_DialogRetryButton"))
        self.ui.cancelMergeInProgress.setIcon(stockIcon("SP_DialogCancelButton"))

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)

    def execute(self, version: Literal["ours", "theirs", "merge", "remerge", "ancestor"]):
        conflict = self.currentConflict

        if not conflict:
            return

        S = GitConflictSides

        if conflict.sides in (S.DeletedByUs, S.AddedByThem):
            # Ours - Keep deletion
            # Theirs - Take incoming changes
            assert conflict.theirs
            assert version in ["ours", "theirs"]

        elif conflict.sides in (S.DeletedByThem, S.AddedByUs):
            # Ours - Ignore deletion
            # Theirs - Take incoming deletion
            assert conflict.ours
            assert version in ["ours", "theirs"]

        elif conflict.sides == S.BothDeleted:
            # Delete the file.
            assert conflict.ancestor
            assert version == "ancestor"
            # Hack: fall back to either side, doesn't matter
            version = "ours"
            assert conflict.ours.path == conflict.ancestor.path
            assert not conflict.ours

        elif conflict.sides in (S.BothModified, S.BothAdded):
            # Pick a side to keep, or merge.
            assert conflict.ours
            assert conflict.theirs
            assert version in ["ours", "theirs", "merge", "remerge"]
            if version in ["merge", "remerge"]:
                reopen = version == "remerge"
                self.openMergeTool(conflict, reopen)

        else:
            raise NotImplementedError(f"unsupported conflict sides: {conflict.sides}")

        if version in ["ours", "theirs"]:
            keepOurs = version == "ours"
            HardSolveConflicts.invoke(self, [self.currentConflict], keepOurs=keepOurs)

    def openMergeTool(self, conflict: GitConflict, reopenWorkInProgress=False):
        OpenMergeTool.invoke(self, conflict, reopenWorkInProgress)
        self.refresh()

    def onMergeDriverResponse(self):
        self.refresh()

    def openPreview(self, oursOrTheirs: Literal["ours", "theirs"]):
        assert oursOrTheirs in ["ours", "theirs"]
        assert self.currentConflict is not None

        if oursOrTheirs == "ours":
            df = self.currentConflict.ours
            prefix = _p("ConflictView", "OUR version")
        else:
            df = self.currentConflict.theirs
            prefix = _p("ConflictView", "THEIR version")

        assert not df.isId0()
        PreviewDeltaFile.invoke(self, df, prefix, self.registerPreviewWindow)

    def registerPreviewWindow(self, preview: CodeWindow):
        self.currentPreviewWindows.append(ReferenceType(preview))
        self.destroyed.connect(preview.close)

    def invalidate(self):
        self.currentConflict = None

        if self.currentMerge is not None:
            self.currentMerge.statusChange.disconnect(self.onMergeDriverResponse)
            self.currentMerge = None

        self.currentMergeState = MergeDriver.State.Idle

        while self.currentPreviewWindows:
            previewWindowRef = self.currentPreviewWindows.pop()
            previewWindow = previewWindowRef()
            if previewWindow is not None:
                previewWindow.close()

    def refresh(self):
        if self.currentConflict is not None:
            self.displayConflict(self.currentConflict, forceRefresh=True)

    def displayConflict(self, conflict: GitConflict, forceRefresh=False):
        assert conflict is not None, "don't call displayConflict with None"

        merge = MergeDriver.findOngoingMerge(conflict)
        state = merge.state if merge else MergeDriver.State.Idle

        # Don't bother refreshing if we're showing the exact same conflict
        if (not forceRefresh
                and conflict == self.currentConflict
                and merge is self.currentMerge
                and state == self.currentMergeState
                and state != MergeDriver.State.Tentative
        ):
            logger.debug("Don't need to refresh ConflictView")
            return

        # If tentative, try to transition to Ready
        if state == MergeDriver.State.Tentative:
            assert merge is not None
            if not merge.checkUnchanged():
                state = MergeDriver.State.Ready
                merge.state = state
            else:
                logger.debug(f"Unchanged: {merge.paths.scratch} {merge.paths.target}")

        self.invalidate()

        self.currentConflict = conflict
        self.currentMerge = merge
        if self.currentMerge:
            self.currentMerge.statusChange.connect(self.onMergeDriverResponse)
            self.currentMergeState = self.currentMerge.state
        else:
            self.currentMergeState = MergeDriver.State.Idle

        sides = conflict.sides
        strings = GitConflictSidesLocalization.getStrings(sides)

        # Reset all text in widgets we can replace placeholder tokens.
        self.ui.retranslateUi(self)

        # Determine page.
        if sides.hasOurs() and sides.hasTheirs():
            page = self.ui.mergePage
        elif sides.hasOurs() or sides.hasTheirs():
            page = self.ui.emptyPage
        else:
            page = self.ui.confirmDeletionPage

        # Hide arrows if all we can do is pick ours/theirs.
        w: QWidget
        for w in self.ui.oursArrow, self.ui.theirsArrow:
            w.setVisible(page is not self.ui.emptyPage)

        # Hide ours/theirs buttons if all we can do is confirm a deletion.
        for w in self.ui.oursButton, self.ui.theirsButton, self.ui.orLabel:
            w.setVisible(page is not self.ui.confirmDeletionPage)

        # Reveal the page
        self.ui.stackedWidget.setCurrentWidget(page)

        self.ui.oursButton.setText(strings.ours1 + "\u2026")
        self.ui.oursButton.setToolTip(strings.ours2)
        self.ui.theirsButton.setText(strings.theirs1 + "\u2026")
        self.ui.theirsButton.setToolTip(strings.theirs2)
        self.ui.explainer.setText(f"<b>{strings.title}.</b> {strings.description}")

        self.ui.oursPreviewButton.setEnabled(sides.hasOurs())
        self.ui.theirsPreviewButton.setEnabled(sides.hasTheirs())

        # Disable ours/theirs buttons while a merge process is running
        self.ui.oursButton.setEnabled(state != MergeDriver.State.Busy)
        self.ui.theirsButton.setEnabled(state != MergeDriver.State.Busy)

        # Ours/theirs status icons
        iconO = "m" if sides.hasOurs() else "missing" if sides == sides.AddedByThem else "d"
        iconT = "m" if sides.hasTheirs() else "missing" if sides == sides.AddedByUs else "d"
        for iconLetter, label in ((iconO, self.ui.oursIcon), (iconT, self.ui.theirsIcon)):
            icon = stockIcon(f"status_{iconLetter}")
            pixmap = icon.pixmap(QSize(16, 16), self.devicePixelRatio())
            label.setPixmap(pixmap)

        # Format placeholders
        displayPath = os.path.basename(self.currentConflict.ours.path)
        formatWidgetText(self.ui.titleLabel, lquo(displayPath))

        tool = lquoe(settings.getMergeToolName())
        for w in self.ui.mergeButton, self.ui.confirmMergeButton, self.ui.discardMergeButton, self.ui.reworkMergeButton:
            formatWidgetText(w, tool=tool)
            formatWidgetTooltip(w, tool=tool)

        # Process debriefing
        if state == MergeDriver.State.Fail:
            self.ui.mergeToolStatus.setText(f"<b style='color: {colors.red.name()}'>{escape(merge.debrief)}</b>")
        else:
            self.ui.mergeToolStatus.setText("")

        # Merge busy/ready
        if state == MergeDriver.State.Busy:
            assert merge is not None
            assert merge.process is not None
            progressMessage = _("Waiting for you to finish merging this file in {0} (PID {1})…",
                                lquoe(merge.processName), merge.process.processId())
            self.ui.mergeInProgressLabel.setText(progressMessage)
            self.ui.stackedWidget.setCurrentWidget(self.ui.mergeInProgressPage)
        elif state == MergeDriver.State.Tentative:
            # Some merge tools like PyCharm may return 0 but postpone writing
            # the actual file from a different process some time later.
            confirmText = _("The file seems unchanged by {0}. Was the merge successful?", lquoe(merge.processName))
            confirmText = f"<b style='color: {colors.red.name()}'>{escape(confirmText)}</b>"
            self.ui.confirmMergeLabel.setText(confirmText)
            self.ui.stackedWidget.setCurrentWidget(self.ui.mergeCompletePage)
        elif state == MergeDriver.State.Ready:
            confirmText = _p("ConflictView", "It looks like you’ve finished merging this file.")
            self.ui.confirmMergeLabel.setText(confirmText)
            self.ui.stackedWidget.setCurrentWidget(self.ui.mergeCompletePage)

    def refreshPrefs(self):
        if self.currentConflict:
            self.refresh()

    def confirmMergeResolution(self):
        merge = self.currentMerge
        assert merge is not None
        self.invalidate()
        AcceptMergeConflictResolution.invoke(self, merge)

    def discardMergeResolution(self):
        merge = self.currentMerge
        assert merge is not None
        merge.deleteNow()
        self.refresh()

    def cancelMergeInProgress(self):
        merge = self.currentMerge
        assert merge is not None
        merge.deleteNow()
        self.refresh()


@dataclass
class GitConflictSidesLocalization:
    description: str
    ours1: str
    ours2: str
    theirs1: str
    theirs2: str
    title: str = "???"

    _cached: ClassVar[dict[GitConflictSides, GitConflictSidesLocalization]] = {}
    _cachedLanguage: ClassVar[str] = ""

    @classmethod
    def getStrings(cls, sides: GitConflictSides) -> GitConflictSidesLocalization:
        table = cls._cached

        if cls._cachedLanguage != settings.prefs.language:
            cls._cachedLanguage = settings.prefs.language
            table.clear()

        try:
            return table[sides]
        except KeyError:
            pass

        table[GitConflictSides.BothModified] = GitConflictSidesLocalization(
            _("This file has received changes from both "
              "<i>our</i> branch and <i>their</i> branch."),
            _("Keep OUR version"),
            _("Keep the file intact in our branch"),
            _("Accept THEIR version"),
            _("Replace the file in our branch with the incoming version"),
        )

        table[GitConflictSides.DeletedByUs] = GitConflictSidesLocalization(
            _("This file was deleted from <i>our</i> branch, "
              "but <i>their</i> branch kept it and made changes to it."),
            _("Keep OUR deletion"),
            _("Don’t resurrect the file in our branch"),
            _("Accept THEIR version"),
            _("Replace the file in our branch with the incoming version"),
        )

        table[GitConflictSides.DeletedByThem] = GitConflictSidesLocalization(
            _("We’ve made changes to this file in <i>our</i> branch, "
              "but <i>their</i> branch has deleted it."),
            _("Keep OUR version"),
            _("Keep the file intact in our branch"),
            _("Accept THEIR deletion"),
            _("Delete the file in our branch"),
        )

        table[GitConflictSides.AddedByUs] = GitConflictSidesLocalization(
            _("No common ancestor."),
            _("Keep OUR version"),
            _("Keep the file intact in our branch"),
            _("Delete it"),
            _("Delete the file in our branch"),
        )

        table[GitConflictSides.AddedByThem] = GitConflictSidesLocalization(
            _("No common ancestor."),
            _("Don’t add"),
            _("Don’t add anything to our branch"),
            _("Accept THEIR version"),
            _("Add the file to our branch"),
        )

        table[GitConflictSides.BothAdded] = GitConflictSidesLocalization(
            _("This file has been created in both <i>our</i> branch "
              "and <i>their</i> branch, independently from each other. "
              "There is no common ancestor."),
            _("Keep OUR version"),
            _("Keep the file intact in our branch"),
            _("Accept THEIR version"),
            _("Replace the file in our branch with the incoming version"),
        )

        table[GitConflictSides.BothDeleted] = GitConflictSidesLocalization(
            _("The file was deleted from <i>our</i> branch, "
              "and <i>their</i> branch has deleted it too."),
            _("Delete it"),
            _("Don’t resurrect the file in our branch"),
            _("Delete it"),
            _("Don’t resurrect the file in our branch"),
        )

        # Fill in titles
        for k, v in table.items():
            v.title = englishTitleCase(trtables.enum(k))

        return table[sides]



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

import re
from collections.abc import Iterable

from gitfourchette import settings
from gitfourchette.graphview.commitlogmodel import CommitLogModel
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID, RepoModel
from gitfourchette.search.itemviewsearchprovider import ItemViewSearchProvider
from gitfourchette.toolbox import *

if TYPE_CHECKING:
    from gitfourchette.graphview.graphview import GraphView


class CommitInfoSearch(ItemViewSearchProvider):
    HashPattern = re.compile(r"[0-9a-fA-F]{1,40}")

    _buddy: GraphView
    likelyHash: bool

    def __init__(self, parent: GraphView):
        super().__init__(parent)
        self.likelyHash = False

    # -------------------------------------------------------------------------
    # ItemViewSearchProvider implementation

    def _walkModelImpl(self, rows: Iterable[int]) -> QModelIndex:
        model = self.buddyModel

        for i in rows:
            index = model.index(i, 0)
            commit = model.data(index, CommitLogModel.Role.Commit)
            if commit is None or commit.id == UC_FAKEID:
                continue
            if self.likelyHash and str(commit.id).startswith(self._term):
                return index
            if self._term in commit.message.lower():
                return index
            if self._term in abbreviatePerson(commit.author, settings.prefs.authorDisplayStyle).lower():
                return index

        raise KeyError()

    # -------------------------------------------------------------------------
    # SearchProvider implementation

    def shortTitle(self) -> str:
        return _("Info")

    def longTitle(self) -> str:
        return _("Find commit hash, message or author")

    def keyboardShortcut(self) -> str:
        return "Alt+I"

    def _termChanged(self):
        super()._termChanged()

        term = self._term
        self.likelyHash = (0 < len(term) <= 40) and bool(self.HashPattern.match(term))

    def notFoundMessage(self) -> str:
        message = super().notFoundMessage()
        return CommitInfoSearch.makeNotFoundMessage(message, self._buddy.repoModel)

    @staticmethod
    def makeNotFoundMessage(message: str, repoModel: RepoModel) -> str:
        limitations = []

        if repoModel.hiddenCommits:
            limitations.append(_("hidden branches"))

        if repoModel.truncatedHistory:
            limitations.append(_("truncated history"))
        elif repoModel.repo.is_shallow:
            limitations.append(_("shallow clone"))

        if limitations:
            message += (f" <span style='color: {mutedToolTipColorHex()}'>" +
                        _("(search limited by {0})", ", ".join(limitations)))

        return message



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

if __name__ == '__main__':
    from gitfourchette.graph import *
    from argparse import ArgumentParser

    parser = ArgumentParser(description="GitFourchette ASCII graph tool")
    parser.add_argument("definition", help="Graph definition (e.g.: \"u:z i:b m:a,b a:z b-c-z\")", nargs="+")
    parser.add_argument("-t", "--tips", nargs="*", default=[])
    parser.add_argument("-x", "--hide", nargs="*", default=[])
    parser.add_argument("-v", "--verbose", action="store_true")
    args = parser.parse_args()

    definition = " ".join(args.definition)
    sequence, heads = GraphDiagram.parseDefinition(definition)

    idSequence = [c.id for c in sequence]
    assert all(c in idSequence for c in args.hide), "one of the given hidden commits isn't in the graph"
    assert all(c in idSequence for c in args.tips), "one of the given tip commits isn't in the graph"

    builder = GraphBuildLoop(args.tips or heads, hideSeeds=args.hide).sendAll(sequence)

    if args.verbose:
        print("Hidden commits:", builder.hiddenCommits)

    diagram = GraphDiagram.diagram(builder.graph, hiddenCommits=builder.hiddenCommits, verbose=args.verbose)
    print(diagram)
